"""deprecated module. Use 'pyunitsystem' instead"""

from tomoscan.utils.io import deprecated_warning

from silx.utils.enum import Enum as _Enum

from pyunitsystem.electriccurrentsystem import (
    ElectricCurrentSystem as _ElectricCurrentSystem,
)
from pyunitsystem.energysystem import EnergySI as _EnergySI
from pyunitsystem.metricsystem import MetricSystem as _MetricSystem
from pyunitsystem.timesystem import TimeSystem as _TimeSystem
from pyunitsystem.unit import Unit as _Unit


class ElectricCurrentSystem(_Enum):
    def __getattribute__(self, name):
        deprecated_warning(
            type_="Module",
            name="tomoscan.unitsystem",
            reason="dedicated project created",
            replacement="pyunitsystem",
            since_version="2.0",
        )
        return _ElectricCurrentSystem.__getattribute__(self, name)


class EnergySI(_Enum):
    def __getattribute__(self, name):
        deprecated_warning(
            type_="Module",
            name="tomoscan.unitsystem",
            reason="dedicated project created",
            replacement="pyunitsystem",
            since_version="2.0",
        )
        return _EnergySI.__getattribute__(self, name)


class MetricSystem(_Enum):
    def __getattribute__(self, name):
        deprecated_warning(
            type_="Module",
            name="tomoscan.unitsystem",
            reason="dedicated project created",
            replacement="pyunitsystem",
            since_version="2.0",
        )
        return _MetricSystem.__getattribute__(self, name)


class TimeSystem(_Enum):
    def __getattribute__(self, name):
        deprecated_warning(
            type_="Module",
            name="tomoscan.unitsystem",
            reason="dedicated project created",
            replacement="pyunitsystem",
            since_version="2.0",
        )
        return _TimeSystem.__getattribute__(self, name)


class Unit(_Enum):
    def __getattribute__(self, name):
        deprecated_warning(
            type_="Module",
            name="tomoscan.unitsystem",
            reason="dedicated project created",
            replacement="pyunitsystem",
            since_version="2.0",
        )
        return _Unit.__getattribute__(self, name)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "01/09/2021"


import os
import shutil
import tempfile
import unittest
from copy import deepcopy

import h5py
import numpy.random
import pytest
from silx.io.url import DataUrl

from tomoscan.scanbase import ReducedFramesInfos, Source, SourceType, TomoScanBase
from tomoscan.test.utils import NXtomoMockContext


class TestFlatFieldCorrection(unittest.TestCase):
    def setUp(self):
        self.data_dir = tempfile.mkdtemp()
        self.scan = TomoScanBase(None, None)
        self.scan.set_reduced_darks(
            {
                0: numpy.random.random(100).reshape((10, 10)),
            }
        )

        self.scan.set_reduced_flats(
            {
                1: numpy.random.random(100).reshape((10, 10)),
                12: numpy.random.random(100).reshape((10, 10)),
                21: numpy.random.random(100).reshape((10, 10)),
            }
        )

        self._data_urls = {}
        projections = {}
        file_path = os.path.join(self.data_dir, "data_file.h5")

        for i in range(-2, 30):
            projections[i] = numpy.random.random(100).reshape((10, 10))
            data_path = "/".join(("data", str(i)))
            self._data_urls[i] = DataUrl(
                file_path=file_path, data_path=data_path, scheme="silx"
            )
            with h5py.File(file_path, mode="a") as h5s:
                h5s[data_path] = projections[i]

        self.scan.projections = projections

    def tearDown(self):
        shutil.rmtree(self.data_dir)

    def test_get_flats_weights(self):
        """test the _get_flats_weights function and insure flat weights
        are correct"""
        flat_weights = self.scan._get_flats_weights()
        self.assertTrue(isinstance(flat_weights, dict))
        self.assertEqual(len(flat_weights), 32)
        self.assertEqual(flat_weights.keys(), self.scan.projections.keys())
        self.assertEqual(flat_weights[-2], {1: 1.0})
        self.assertEqual(flat_weights[0], {1: 1.0})
        self.assertEqual(flat_weights[1], {1: 1.0})
        self.assertEqual(flat_weights[12], {12: 1.0})
        self.assertEqual(flat_weights[21], {21: 1.0})
        self.assertEqual(flat_weights[24], {21: 1.0})

        def assertAlmostEqual(ddict1, ddict2):
            self.assertEqual(ddict1.keys(), ddict2.keys())
            for key in ddict1.keys():
                self.assertAlmostEqual(ddict1[key], ddict2[key])

        assertAlmostEqual(flat_weights[2], {1: 10.0 / 11.0, 12: 1.0 / 11.0})
        assertAlmostEqual(flat_weights[10], {1: 2.0 / 11.0, 12: 9.0 / 11.0})
        assertAlmostEqual(flat_weights[18], {12: 3.0 / 9.0, 21: 6.0 / 9.0})

    def test_flat_field_data_url(self):
        """insure the flat_field is computed. Simple processing test when
        provided data is a DataUrl"""
        projections_keys = [key for key in self.scan.projections.keys()]
        projections_urls = [self.scan.projections[key] for key in projections_keys]
        self.scan.flat_field_correction(projections_urls, projections_keys)

    def test_flat_field_data_numpy_array(self):
        """insure the flat_field is computed. Simple processing test when
        provided data is a numpy array"""
        self.scan.projections = self._data_urls
        projections_keys = [key for key in self.scan.projections.keys()]
        projections_urls = [self.scan.projections[key] for key in projections_keys]
        self.scan.flat_field_correction(projections_urls, projections_keys)


def test_Source_API():
    """Test Source API"""
    source = Source(name="my source", type=SourceType.SYNCHROTRON_X_RAY_SOURCE)
    source.name = "toto"
    with pytest.raises(TypeError):
        source.name = 12
    assert isinstance(source.name, str)
    source.type = SourceType.FREE_ELECTRON_LASER
    assert isinstance(source.type, SourceType)
    source.type = None
    str(source)


def test_TomoScanBase_API():
    """Test TomoScanBase API"""
    with pytest.raises(NotImplementedError):
        TomoScanBase.is_tomoscan_dir("")
    with pytest.raises(NotImplementedError):
        TomoScanBase(scan="", type_="undefined").is_abort()
    scan = TomoScanBase(scan="", type_="undefined")
    scan.source
    scan.flats = {1: numpy.random.random(100 * 100).reshape(100, 100)}
    assert len(scan.flats) == 1

    scan.darks = {0: numpy.random.random(100 * 100).reshape(100, 100)}
    assert len(scan.darks) == 1

    scan.alignment_projections = {
        2: numpy.random.random(100 * 100).reshape(100, 100),
        3: numpy.random.random(100 * 100).reshape(100, 100),
    }
    assert len(scan.alignment_projections) == 2

    for prop in (
        "dark_n",
        "tomo_n",
        "flat_n",
        "pixel_size",
        "instrument_name",
        "dim_1",
        "dim_2",
        "scan_range",
        "ff_interval",
        "energy",
        "intensity_monitor",
        "field_of_view",
        "estimated_cor_frm_motor",
        "x_translation",
        "y_translation",
        "z_translation",
        "sequence_name",
        "sample_name",
        "group_size",
    ):
        with pytest.raises(NotImplementedError):
            getattr(scan, prop)

    assert isinstance(scan.to_dict(), dict)

    for fct in (
        "update",
        "get_proj_angle_url",
        "get_projections_intensity_monitor",
        "get_flat_expected_location",
        "get_dark_expected_location",
        "get_projection_expected_location",
        "get_energy_expected_location",
        "get_distance_expected_location",
        "get_pixel_size_expected_location",
    ):
        with pytest.raises(NotImplementedError):
            getattr(scan, fct)()


def test_save_load_reduced_darks(tmpdir):
    with NXtomoMockContext(
        scan_path=os.path.join(tmpdir, "test_save_load_reduced_darks"),
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
    ) as scan:
        with pytest.raises(TypeError):
            scan.save_reduced_darks(
                darks=None,
                output_urls=(scan.REDUCED_DARKS_DATAURLS,),
            )

        with pytest.raises(TypeError):
            scan.save_reduced_darks(
                darks={
                    0: numpy.ones((10, 10)),
                },
                output_urls=None,
            )

        scan.path = None
        with pytest.raises(ValueError):
            scan.save_reduced_darks(
                darks={
                    0: numpy.ones((10, 10)),
                },
                output_urls=(scan.REDUCED_DARKS_DATAURLS,),
            )

        with pytest.raises(ValueError):
            scan.save_reduced_darks(
                darks={
                    0: numpy.ones((10, 10)),
                },
                output_urls=(scan.REDUCED_DARKS_DATAURLS,),
            )


def test_ReducedFramesInfo():
    """
    test ReducedFramesInfos class
    """
    infos = ReducedFramesInfos()
    assert infos.to_dict() == {}
    infos.count_time = numpy.array([12.3, 13.0])
    assert infos.count_time == [12.3, 13.0]
    infos.machine_electric_current = [23.5, 56.9]
    assert infos.machine_electric_current == [23.5, 56.9]
    my_dict = deepcopy(infos.to_dict())
    assert my_dict == {
        ReducedFramesInfos.COUNT_TIME_KEY: [12.3, 13.0],
        ReducedFramesInfos.MACHINE_ELECT_CURRENT_KEY: [23.5, 56.9],
    }

    infos.clear()
    assert infos.to_dict() == {}
    new_infos = ReducedFramesInfos()
    new_infos.load_from_dict(my_dict)
    assert new_infos.to_dict() == my_dict

    with pytest.raises(TypeError):
        new_infos.count_time = 12
    with pytest.raises(TypeError):
        new_infos.machine_electric_current = 12

    str(new_infos)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/07/2022"


import os

from tomoscan.identifier import VolumeIdentifier
from tomoscan.utils import docstring


class RawVolumeIdentifier(VolumeIdentifier):
    """Identifier for the .vol volume"""

    def __init__(self, object, file_path):
        super().__init__(object)
        self._file_path = os.path.realpath(os.path.abspath(file_path))

    @docstring(VolumeIdentifier)
    def short_description(self) -> str:
        return f"{self.scheme}:{self.tomo_type}:{os.path.basename(self._file_path)}"

    @property
    def file_path(self):
        return self._file_path

    @property
    @docstring(VolumeIdentifier)
    def scheme(self) -> str:
        return "raw"

    def __str__(self):
        return f"{self.scheme}:{self.tomo_type}:{self._file_path}"

    def __eq__(self, other):
        if isinstance(other, RawVolumeIdentifier):
            return (
                self.tomo_type == other.tomo_type
                and self._file_path == other._file_path
            )
        else:
            return False

    def __hash__(self):
        return hash(self._file_path)

    @staticmethod
    def from_str(identifier):
        identifier_no_scheme = identifier.split(":")[-1]
        vol_file = identifier_no_scheme
        from tomoscan.esrf.volume.rawvolume import RawVolume

        return RawVolumeIdentifier(object=RawVolume, file_path=vol_file)



from __future__ import annotations
import logging
import os
import shutil
from urllib.request import ProxyHandler, build_opener, urlopen

try:
    import gitlab
except ImportError:
    __has_gitlab__ = False
else:
    __has_gitlab__ = True


_logger = logging.getLogger(__name__)


def wget_file(url, output, timeout=1000):
    """
    straight wget on a file - required for LFS files
    """
    _logger.info(f"Trying to download scan {url}, timeout set to {timeout}s")
    dictProxies = {}
    if "http_proxy" in os.environ:
        dictProxies["http"] = os.environ["http_proxy"]
        dictProxies["https"] = os.environ["http_proxy"]
    if "https_proxy" in os.environ:
        dictProxies["https"] = os.environ["https_proxy"]
    if dictProxies:
        proxy_handler = ProxyHandler(dictProxies)
        opener = build_opener(proxy_handler).open
    else:
        opener = urlopen
    _logger.info(f"wget {url}")
    data = opener(url, data=None, timeout=timeout).read()
    _logger.info(f"{url} successfully downloaded.")

    try:
        with open(output, "wb") as outfile:
            outfile.write(data)
    except IOError as e:
        raise IOError(
            f"unable to write downloaded data to disk at {output}. Error is {e}"
        )


def download_file(
    host: str,
    token: str | None,
    project_id: int,
    branch_name: str,
    file_path: str,
    output: str,
    use_lfs=True,
):
    """
    if the download / REST API fails the backup is to do a git clone of the project
    with the requested branch under nxtomomill/test/utils/__archive__
    """
    if not __has_gitlab__:
        raise ImportError(
            "gitlab not install. Did you install the 'test' extra requirements ? "
        )

    try:
        gl = gitlab.Gitlab(host, private_token=token)
        project = gl.projects.get(project_id)
    except Exception as e:
        raise RuntimeError(f"Failed to download {file_path}. Error is {e}")
    else:
        # if needed create output folder
        os.makedirs(os.path.dirname(output), exist_ok=True)
        if use_lfs:
            # don't know why but for now the REST API does not return the information about the file.
            # so go trhough get for now...
            project_web_url = project.web_url
            wget_file(
                url=f"{project_web_url}/-/raw/{branch_name}/{file_path}?ref_type=heads",
                output=output,
            )
        else:
            # save the file
            with open(output, "wb") as f:
                project.files.raw(
                    file_path=file_path,
                    ref=branch_name,
                    streamed=True,
                    action=f.write,
                )
    return output


def download_folder(
    host: str,
    token: str | None,
    project_id: int,
    branch_name: str,
    folder: str,
    output: str,
    overwrite: bool = True,
    use_lfs=True,
):
    """

    :param bool overwrite: if the file already exists overwrite it
    if the download / REST API fails the backup is to do a git clone of the project
    with the requested branch under nxtomomill/test/utils/__archive__
    """
    if not __has_gitlab__:
        raise ImportError(
            "gitlab not install. Did you install the 'test' extra requirements ? "
        )

    gl = gitlab.Gitlab(host, private_token=token)
    project = gl.projects.get(project_id)
    tree = project.repository_tree(
        path=folder,
        ref=branch_name,
        get_all=True,  # if false not all items will be return
    )
    for item in tree:
        if item.get("type") == "tree":
            # if this is a folder
            download_folder(
                host=host,
                token=token,
                project_id=project_id,
                branch_name=branch_name,
                folder=item.get("path"),
                output=os.path.join(output, item.get("path")),
                overwrite=overwrite,
                use_lfs=use_lfs,
            )
        elif not overwrite and os.path.exists(os.path.join(output, item.get("path"))):
            continue
        else:
            download_file(
                host=host,
                token=token,
                project_id=project_id,
                branch_name=branch_name,
                file_path=item.get("path"),
                output=os.path.join(output, item.get("path")),
                use_lfs=use_lfs,
            )


class GitlabProject:
    """
    simple manager to download a file or a folder from a gitlab project.
    If the project is not public then a token must be provided.

    All downloaded files will be stored in __cache__ folder.
    If the file is already in the __cache__ then it will be picked drectly from there.
    """

    def __init__(
        self, host, project_id, cache_dir, token=None, branch_name="main"
    ) -> None:
        self._host = host
        self._project_id = project_id
        self._token = token
        self._branch_name = branch_name
        self._cache_dir = cache_dir

    @property
    def host(self) -> str:
        return self._host

    @property
    def project_id(self) -> int:
        return self._project_id

    @property
    def token(self):
        return self._token

    @property
    def branch_name(self) -> str:
        return self._branch_name

    @property
    def cache_dir(self) -> str:
        return self._cache_dir

    def clear_cache(self):
        if os.path.exists(self.cache_dir):
            shutil.rmtree(self.cache_dir)

    def get_dataset(self, name, is_folder=False, use_lfs=True) -> str | None:
        """
        download dataset. Name is the path to the dataset.
        for now user must notify if the dataset is a folder or not
        I guess we can get this information from gitlab and the REST API
        but to be done latter as an improvement
        """
        file_archive_location = os.path.join(self.cache_dir, name)

        if is_folder:
            try:
                download_folder(
                    host=self.host,
                    token=self.token,
                    project_id=self.project_id,
                    branch_name=self.branch_name,
                    folder=name,
                    output=self.cache_dir,
                    overwrite=False,
                    use_lfs=use_lfs,
                )
            except Exception as e:
                _logger.error(str(e))
                return None
            else:
                return file_archive_location
        elif os.path.exists(file_archive_location):
            # if exists, avoid redownloading it
            return file_archive_location
        else:
            try:
                download_file(
                    host=self.host,
                    token=self.token,
                    project_id=self.project_id,
                    branch_name=self.branch_name,
                    file_path=name,
                    output=file_archive_location,
                    use_lfs=use_lfs,
                )
            except Exception as e:
                _logger.error(str(e))
                return None
            else:
                return file_archive_location


GitlabDataset = GitlabProject(
    branch_name="tomoscan",
    host="https://gitlab.esrf.fr",
    cache_dir=os.path.join(
        os.path.dirname(__file__),
        "__archive__",
    ),
    token=None,
    project_id=4299,  # https://gitlab.esrf.fr/tomotools/ci_datasets
)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""module to define base class for a volume"""


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "27/01/2022"


from typing import Optional, Union
import logging
import numpy
from silx.io.url import DataUrl
from silx.math.combo import (  # pylint: disable=E0611 (not found from the pyx file)
    min_max,
)
from pyunitsystem.metricsystem import MetricSystem

from tomoscan.utils.io import deprecated, deprecated_warning
from tomoscan.identifier import VolumeIdentifier
from tomoscan.scanbase import TomoScanBase
from tomoscan.tomoobject import TomoObject
from tomoscan.utils import BoundingBox1D, BoundingBox3D

_logger = logging.getLogger(__name__)


class VolumeBase(TomoObject):
    """
    context: we aim at having a common way of saving and loading volumes through the tomotools suite.
    The goal is to aim handling of volumes when creating them or doing some operations with those like stitching...

    :param DataUlr url: url of the volume. Could be path to a master file if we can provide one per each volume. Otherwise could be a pattern of edf files or tiff file with a data range
    :param Optional[TomoScanBase] source_scan: potential instance of TomoScanBase in order to get extra information. This could be saved in the volume file to (external link)
    :param Optional[nump.ndarray] data: volume data. Expected to be 3D
    :param Optional[dict] metadata: metadata associated to the volume. Must be a dict of serializable object
    :param Optional[DataUrl] data_url: url to save the data. If provided url must not be provided. If an object is constructed from data and metadta url then no rule to create a VolumeIdentifier can be created and call to et_identifier will raise an error.
    :param Optional[DataUrl] metadata_url: url to save the metadata. If provided url must not be provided. If an object is constructed from data and metadta url then no rule to create a VolumeIdentifier can be created and call to et_identifier will raise an error.
    :param bool overwrite: when save the data if encounter a ressource already existing overwrite it (if True) or not.
    :param str overwrite: when save the data if encounter a ressource already existing overwrite it (if True) or not.

    :raises TypeError:
    :raises ValueError: * if data is a numpy array and not 3D.
    :raises OSError:
    """

    EXTENSION = None

    def __init__(
        self,
        url: Optional[DataUrl] = None,
        data: Optional[numpy.ndarray] = None,
        source_scan: Optional[TomoScanBase] = None,
        metadata: Optional[dict] = None,
        data_url: Optional[DataUrl] = None,
        metadata_url: Optional[DataUrl] = None,
        overwrite: bool = False,
        data_extension: Optional[str] = None,
        metadata_extension: Optional[str] = None,
    ) -> None:
        super().__init__()
        if url is not None and (data_url is not None or metadata_url is not None):
            raise ValueError(
                "Either url or (data_url and / or metadata_url) can be provided not both"
            )

        # warning on source_scan: should be defined before the url because deduce_data_and_metadata_urls might need it
        # Then as source scan can imply several modification of url... we can only define it during construction and this
        # must not involve with object life
        if not isinstance(source_scan, (TomoScanBase, type(None))):
            raise TypeError(
                f"source scan is expected to be None or an instance of TomoScanBase. Not {type(source_scan)}"
            )
        self.__source_scan = source_scan
        self._data_extension = data_extension
        self._metadata_extension = metadata_extension

        self.overwrite = overwrite
        self.url = url
        self.metadata = metadata
        self.data = data

        if url is None:
            self._data_url = data_url
            self._metadata_url = metadata_url
        else:
            # otherwise have be setted when url has been set from call to deduce_data_and_metadata_urls
            pass

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url: Optional[DataUrl]) -> None:
        if url is not None and not isinstance(url, DataUrl):
            raise TypeError
        self._url = url
        self._data_url, self._metadata_url = self.deduce_data_and_metadata_urls(url)

    def deduce_data_and_metadata_urls(self, url: Optional[DataUrl]) -> tuple:
        """
        compute data and metadata urls from 'parent url'
        :return: data_url: Optional[DataUrl], metadata_url: Optional[DataUrl]
        """
        raise NotImplementedError("Base class")

    @property
    def data_extension(self):
        return self._data_extension

    @property
    def metadata_extension(self):
        return self._metadata_extension

    @property
    def data_url(self):
        return self._data_url

    @property
    def metadata_url(self):
        return self._metadata_url

    @property
    def data(self) -> Optional[numpy.ndarray]:
        return self._data

    @data.setter
    def data(self, data):
        if not isinstance(data, (numpy.ndarray, type(None))):
            raise TypeError(
                f"data is expected to be None or a numpy array not {type(data)}"
            )
        if isinstance(data, numpy.ndarray) and data.ndim != 3:
            raise ValueError(f"data is expected to be 3D and not {data.ndim}D.")
        self._data = data

    def get_slice(
        self,
        index: Union[str, int] = None,
        axis=None,
        xy=None,
        xz=None,
        yz=None,
        url: Optional[DataUrl] = None,
    ):
        if xy is yz is xz is axis is None:
            raise ValueError("axis should be provided")

        if self.data is None:
            # fixme: must be redefined by inheriting classes.
            # for example for single base frame we are simply loading the full volume instead of retrieving the
            # file. This is a bottleneck especially for xy slice because all the files are loaded instead of one
            # in the worst case.
            self.load_data(url=url, store=True)

        if isinstance(index, str):
            if index == "middle":
                index = self.get_volume_shape()[axis] // 2
            elif index == "first":
                index = 0
            elif index == "last":
                index = -1

        if self.data is not None:
            return self.select(
                volume=self.data, xy=xy, xz=xz, yz=yz, axis=axis, index=index
            )
        else:
            return None

    @property
    def metadata(self) -> Optional[dict]:
        return self._metadata

    @metadata.setter
    def metadata(self, metadata: Optional[dict]):
        if not isinstance(metadata, (dict, type(None))):
            raise TypeError(
                f"metadata is expected to be None or a dict not {type(metadata)}"
            )
        self._metadata = metadata

    @staticmethod
    def example_defined_from_str_identifier() -> str:
        """example as string to explain how users can defined identifiers from a string"""
        raise NotImplementedError("Base class")

    def clear_cache(self):
        """remove object stored in data and medatada"""
        self.data = None
        self.metadata = None

    # generic function requested
    @property
    def source_scan(self) -> Optional[TomoScanBase]:
        return self.__source_scan

    @property
    def overwrite(self) -> bool:
        return self._overwrite

    @overwrite.setter
    def overwrite(self, overwrite: bool) -> None:
        if not isinstance(overwrite, bool):
            raise TypeError
        self._overwrite = overwrite

    # function to be loaded to an url
    @staticmethod
    def from_identifier(identifier: Union[str, VolumeIdentifier]):
        """Return the Dataset from a identifier"""
        raise NotImplementedError("Base class")

    def get_identifier(self) -> VolumeIdentifier:
        """dataset unique identifier. Can be for example a hdf5 and
        en entry from which the dataset can be rebuild"""
        raise NotImplementedError("Base class")

    # utils required for operations like stitching

    @staticmethod
    def _insure_reconstruction_dict_exists(ddict):
        if "processing_options" not in ddict:
            ddict["processing_options"] = {}
        if "reconstruction" not in ddict["processing_options"]:
            ddict["processing_options"]["reconstruction"] = {}

    @property
    def position(self) -> Optional[tuple]:
        """position are provided as a tuple using the same reference for axis as the volume data.
        position is returned as (axis_0_pos, axis_1_pos, axis_2_pos). Can also be see as (z_position, y_position, x_position)
        """
        metadata = self.metadata or self.load_metadata()
        position = (
            metadata.get("processing_options", {})
            .get("reconstruction", {})
            .get("position", None)
        )
        if position is None:
            return None
        else:
            return tuple(position)

    @position.setter
    def position(self, position) -> None:
        if self.metadata is None:
            self.metadata = {}
        self._insure_reconstruction_dict_exists(self.metadata)
        self.metadata["processing_options"]["reconstruction"]["position"] = numpy.array(
            position
        )

    @property
    @deprecated(replacement="voxel_size", since_version="1.3.0")
    def pixel_size(self):
        return self.voxel_size

    @pixel_size.setter
    @deprecated(replacement="voxel_size", since_version="1.3.0")
    def pixel_size(self, pixel_size) -> None:
        if numpy.isscalar(pixel_size):
            pixel_size = [pixel_size] * 3
            _logger.warning(
                "pixel_size is expected to be a tuple. Conversion will be removed soon. Please update your code"
            )
        self.voxel_size = pixel_size

    @property
    def voxel_size(self) -> Optional[tuple]:
        """
        voxel size as (axis 0 dim - aka z, axis 1 dim - aka y, axis 2 dim aka z)
        """
        metadata = self.metadata or self.load_metadata()

        voxel_size = (
            metadata.get("processing_options", {})
            .get("reconstruction", {})
            .get("voxel_size_cm", None)
        )
        if voxel_size is None:
            # ensure backward compatibility with old volumes (before nabu 2023.1)
            # try fall back on pixel_size_cm for old volumes
            voxel_size = (
                metadata.get("processing_options", {})
                .get("reconstruction", {})
                .get("pixel_size_cm", None)
            )
            if voxel_size is not None:
                voxel_size = (voxel_size, voxel_size, voxel_size)

        if voxel_size is not None:
            if numpy.isscalar(voxel_size):
                # be safer and handle the case voxel size is a scalar
                voxel_size = [voxel_size] * 3
            return tuple(numpy.array(voxel_size) * MetricSystem.CENTIMETER.value)
        return None

    @voxel_size.setter
    def voxel_size(self, voxel_size: tuple) -> None:
        if self.metadata is None:
            self.metadata = {}
        if numpy.isscalar(voxel_size):
            raise TypeError("voxel is expected to be a tuple of three values")
        self._insure_reconstruction_dict_exists(self.metadata)
        self.metadata["processing_options"]["reconstruction"]["voxel_size_cm"] = (
            numpy.array(voxel_size) / MetricSystem.CENTIMETER.value
        )

    def get_volume_shape(self, url=None):
        raise NotImplementedError("Base class")

    def get_bounding_box(self, axis: Optional[Union[str, int]] = None):
        if axis is None:
            x_bb = self.get_bounding_box(axis=2)
            y_bb = self.get_bounding_box(axis=1)
            z_bb = self.get_bounding_box(axis=0)
            return BoundingBox3D(
                (z_bb.min, y_bb.min, x_bb.min),
                (z_bb.max, y_bb.max, x_bb.max),
            )

        position = self.position
        shape = self.get_volume_shape()
        # TODO: does it make sense that pixel size is a scalar ?
        voxel_size = self.voxel_size
        missing = []
        if position is None:
            missing.append("position")
        if shape is None:
            missing.append("shape")
            raise ValueError("Unable to find volume shape")
        if voxel_size is None:
            missing.append("pixel_size")

        if len(missing) > 0:
            raise ValueError(
                f"Unable to get bounding box. Missing information: {'; '.join(missing)}"
            )
        assert axis is not None
        if isinstance(axis, str):
            if axis == "x":
                axis = 2
            elif axis == "y":
                axis = 1
            elif axis == "z":
                axis = 0
            else:
                raise ValueError(f"axis '{axis}' is not handled")
        min_pos_in_meter = position[axis] - voxel_size[axis] * shape[axis] / 2.0
        max_pos_in_meter = position[axis] + voxel_size[axis] * shape[axis] / 2.0
        return BoundingBox1D(min_pos_in_meter, max_pos_in_meter)

    def get_min_max(self) -> tuple:
        """
        compute min max of the volume. Can take some time but avoid to load the full volume in memory
        """
        if self.data is not None:
            return self.data.min(), self.data.max()
        else:
            min_v, max_v = None, None
            for s in self.browse_slices():
                min_v = min(min_v, s.min()) if min_v is not None else s.min()
                max_v = max(max_v, s.max()) if max_v is not None else s.max()
            return min_v, max_v

    # load / save stuff

    @property
    def extension(self) -> str:
        return self.EXTENSION

    def load(self):
        self.load_metadata(store=True)
        # always load metadata first because we might expect to get some information from
        # it in order to load data next
        self.load_data(store=True)

    def save(self, url: Optional[DataUrl] = None, **kwargs):
        if url is not None:
            data_url, metadata_url = self.deduce_data_and_metadata_urls(url=url)
        else:
            data_url = self.data_url
            metadata_url = self.metadata_url
        self.save_data(data_url, **kwargs)
        if self.metadata is not None:
            # a volume is not force to have metadata to save. But calling save_metadata direclty might raise an error
            # if no metadata found
            self.save_metadata(metadata_url)

    def save_data(self, url: Optional[DataUrl] = None, **kwargs) -> None:
        """
        save data to the provided url or existing one if none is provided
        """
        raise NotImplementedError("Base class")

    def save_metadata(self, url: Optional[DataUrl] = None) -> None:
        """
        save metadata to the provided url or existing one if none is provided
        """
        raise NotImplementedError("Base class")

    def load_data(
        self, url: Optional[DataUrl] = None, store: bool = True
    ) -> numpy.ndarray:
        raise NotImplementedError("Base class")

    def load_metadata(self, url: Optional[DataUrl] = None, store: bool = True) -> dict:
        raise NotImplementedError("Base class")

    def check_can_provide_identifier(self):
        if self.url is None:
            raise ValueError(
                "Unable to provide an identifier. No url has been provided"
            )

    @staticmethod
    def select(volume, xy=None, xz=None, yz=None, axis=None, index=None):
        if xy is not None:
            deprecated_warning(
                type_="parameter",
                name="xy",
                replacement="axis and index",
            )
            if axis is None and index is None:
                axis = 0
                index = xy
            else:
                raise ValueError("several axis (previously xy, xz, yz requested")
        elif xz is not None:
            deprecated_warning(
                type_="parameter",
                name="xz",
                replacement="axis and index",
            )
            if axis is None and index is None:
                axis = 1
                index = xz
            else:
                raise ValueError("several axis (previously xy, xz, yz requested")
        elif yz is not None:
            deprecated_warning(
                type_="parameter",
                name="yz",
                replacement="axis and index",
            )
            if axis is None and index is None:
                axis = 2
                index = yz
            else:
                raise ValueError("several axis (previously xy, xz, yz requested")

        if not volume.ndim == 3:
            raise TypeError(f"volume is expected to be 3D. {volume.ndim}D provided")
        if axis == 0:
            return volume[index]
        elif axis == 1:
            return volume[:, index]
        elif axis == 2:
            return volume[:, :, index]
        else:
            raise ValueError(f"axis {axis} is not handled")

    def browse_data_files(self, url=None):
        """
        :param url: data url. If not provided will take self.data_url

        return a generator go through all the existings files associated to the data volume
        """
        raise NotImplementedError("Base class")

    def browse_metadata_files(self, url=None):
        """
        :param url: metadata url. If not provided will take self.metadata_url

        return a generator go through all the existings files associated to the data volume
        """
        raise NotImplementedError("Base class")

    def browse_data_urls(self, url=None):
        """
        generator on data urls used.

        :param url: data url to be used. If not provided will take self.data_url
        """
        raise NotImplementedError("Base class")

    def browse_slices(self, url=None):
        """
        generator of 2D numpy array representing a slice

        :param url: data url to be used. If not provided will browse self.data if exists else self.data_url
        :warning: this will get the slice from the data on disk and never use `data` property.
                  so before browsing slices you might want to check if data is already loaded
        """
        raise NotImplementedError("Base class")

    def load_chunk(self, chunk, url=None):
        """
        Load a sub-volume.

        :param chunk: tuple of slice objects indicating which chunk of the volume has to be loaded.
        :param url: data url to be used. If not provided will take self.data_url
        """
        raise NotImplementedError("Base class")

    def get_min_max_values(self, url=None) -> tuple:
        """
        compute min max over 'data' if exists else browsing the volume slice by slice

        :param url: data url to be used. If not provided will take self.data_url
        """
        min_v = None
        max_v = None
        if self.data is not None:
            data = self.data
        else:
            data = self.browse_slices(url=url)
        for data_slice in data:
            if min_v is None:
                min_v = data_slice.min()
                max_v = data_slice.max()
            else:
                min_lv, max_lv = min_max(data_slice, finite=True)
                min_v = min(min_v, min_lv)
                max_v = max(max_v, max_lv)
        return min_v, max_v

    def data_file_saver_generator(self, n_frames, data_url: DataUrl, overwrite: bool):
        """
        Provide a helper class to dump data frame by frame. For know the only possible interaction is
        Helper[:] = frame

        :param int n_frames: number of frame the final volume will contain
        :param DataUrl data_url: url to dump data
        :param bool overwrite: overwrite existing file ?
        """
        raise NotImplementedError("Base class")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""module to mock volume"""


from typing import Sized, Union

import numpy
from silx.image.phantomgenerator import PhantomGenerator
from silx.utils.enum import Enum as _Enum


class Scene(_Enum):
    SHEPP_LOGAN = "Shepp-Logan"


def create_volume(
    frame_dims: Union[int, tuple], z_size: int, scene: Scene = Scene.SHEPP_LOGAN
) -> numpy.ndarray:
    """
    create a numpy array of the requested schene for a total of frames_dimes*z_size elements

    :param tuple frame_dims: 2d tuple of frame dimensions
    :param int z_size: number of elements on the volume on z axis
    :param Scene scene: scene to compose
    """
    scene = Scene.from_value(scene)
    if not isinstance(z_size, int):
        raise TypeError(
            f"z_size is expected to be an instance of int not {type(z_size)}"
        )
    if scene is Scene.SHEPP_LOGAN:
        if isinstance(frame_dims, Sized):
            if not len(frame_dims) == 2:
                raise ValueError(
                    f"frame_dims is expected to be an integer or a list of two integers. Not {frame_dims}"
                )
            if frame_dims[0] != frame_dims[1]:
                raise ValueError(
                    f"{scene} only handle square frame. Frame width and height should be the same"
                )
            else:
                dim = frame_dims[0]
        elif isinstance(frame_dims, int):
            dim = frame_dims
        else:
            raise TypeError(
                f"frame_dims is expected to be a list of two integers or an integer. Not {frame_dims}"
            )
        return numpy.asarray(
            [PhantomGenerator.get2DPhantomSheppLogan(dim) * 10000.0] * z_size
        )
    else:
        raise NotImplementedError



"""module dedicated to nexus - NXtomo - scan class (NXtomoScan) and utils. Based on HDF5 file format"""


__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "09/08/2018"


import io
import json
import logging
import os
import typing

import h5py
import numpy
from silx.io.url import DataUrl
from silx.io.utils import get_data, h5py_read_dataset

from pyunitsystem import electriccurrentsystem, energysystem, timesystem
from pyunitsystem.metricsystem import MetricSystem
from pyunitsystem.unit import Unit

from nxtomo.paths.nxtomo import get_paths as _get_nexus_paths
from nxtomo.utils.transformation import (
    get_lr_flip,
    get_ud_flip,
    LRDetTransformation,
    UDDetTransformation,
)
from nxtomo.nxobject.nxdetector import NXdetector
from nxtomo.paths.nxtomo import LATEST_VERSION as LATEST_NXTOMO_VERSION
from nxtomo.application.nxtomo import ImageKey, NXtomo as _NXtomo

from tomoscan.utils.io import deprecated, deprecated_warning
from tomoscan.esrf.identifier.hdf5Identifier import NXtomoScanIdentifier
from tomoscan.esrf.scan.framereducer.hdf5framereducer import HDF5FrameReducer
from tomoscan.identifier import ScanIdentifier
from tomoscan.io import HDF5File
from tomoscan.scanbase import FOV, Source, TomoScanBase
from tomoscan.utils import BoundingBox1D, BoundingBox3D, docstring
from tomoscan.esrf.scan.utils import (
    dataset_has_broken_vds,
    check_possible_issue_with_rlimit,
)


from .utils import get_compacted_dataslices

_logger = logging.getLogger(__name__)


@deprecated(reason="moved", replacement="nxtomo.paths.nxtomo", since_version="0.8.0")
def get_nexus_paths(version: float):
    return _get_nexus_paths(version=version)


class NXtomoScan(TomoScanBase):
    """
    This is the implementation of a TomoBase class for an acquisition stored
    in a HDF5 file. HDF5 may have broken links in virtual layout, considering
    this the class performs an integrity check on all the links and this
    can be very long for huge data sets with tens of thousands of links.
    This check can be set by defining and exportin the shell environment variable
    TOMOTOOLS_SKIP_DET_CHECK  setted either to 1 or y.

    For now several property of the acquisition is accessible thought a getter
    (like get_scan_range) and a property (scan_range).

    This is done to be compliant with TomoBase instantiation. But his will be
    replace progressively by properties at the 'TomoBase' level

    :param scan: scan directory or scan masterfile.h5
    :param Union[str, None] entry: name of the NXtomo entry to select. If given
                                   index is ignored.
    :param Union[int, None] index: of the NXtomo entry to select. Ignored if
                                   an entry is specified. For consistency
                                   entries are ordered alphabetically
    :param Union[float, None] nx_version: Version of the Nexus convention to use.
                                          By default (None) it will take the latest one
    """

    _NEXUS_VERSION_PATH = "version"

    _TYPE = "hdf5"

    _DICT_ENTRY_KEY = "entry"

    SCHEME = "silx"

    REDUCED_DARKS_DATAURLS = (
        DataUrl(
            file_path="{scan_prefix}_darks.hdf5",
            data_path="{entry}/darks/{index}",
            scheme=SCHEME,
        ),
    )

    REDUCED_DARKS_METADATAURLS = (
        DataUrl(
            file_path="{scan_prefix}_darks.hdf5",
            data_path="{entry}/darks/",
            scheme=SCHEME,
        ),
    )

    REDUCED_FLATS_DATAURLS = (
        DataUrl(
            file_path="{scan_prefix}_flats.hdf5",
            data_path="{entry}/flats/{index}",
            scheme=SCHEME,
        ),
    )

    REDUCED_FLATS_METADATAURLS = (
        DataUrl(
            file_path="{scan_prefix}_flats.hdf5",
            data_path="{entry}/flats/",
            scheme=SCHEME,
        ),
    )

    FRAME_REDUCER_CLASS = HDF5FrameReducer

    def __init__(
        self,
        scan: str,
        entry: str = None,
        index: typing.Optional[int] = 0,
        ignore_projections: typing.Optional[typing.Iterable] = None,
        nx_version=None,
    ):
        if entry is not None:
            index = None
        # if the user give the master file instead of the scan dir...
        if scan is not None:
            if not os.path.exists(scan) and "." in os.path.split(scan)[-1]:
                self.master_file = scan
                scan = os.path.dirname(scan)
            elif os.path.isfile(scan) or ():
                self.master_file = scan
                scan = os.path.dirname(scan)
            else:
                self.master_file = self.get_master_file(scan)
        else:
            self.master_file = None

        super(NXtomoScan, self).__init__(
            scan=scan, type_=NXtomoScan._TYPE, ignore_projections=ignore_projections
        )

        if scan is None:
            self._entry = None
        else:
            self._entry = entry or self._get_entry_at(
                index=index, file_path=self.master_file
            )
            if self._entry is None:
                raise ValueError(f"unable to find a valid entry for {self.master_file}")
        # for now the default entry is 1_tomo but should change with time
        self._name = None
        self._sample_name = None
        self._grp_size = None
        # data caches
        self._projections_compacted = None
        self._flats = None
        self._darks = None
        self._tomo_n = None
        # number of projections / radios
        self._dark_n = None
        # number of dark image made during acquisition
        self._flat_n = None
        # number of flat field made during acquisition
        self._scan_range = None
        # scan range, in degree
        self._dim_1, self._dim_2 = None, None
        # image dimensions
        self._x_pixel_size = None
        self._y_pixel_size = None
        # pixel dimensions (tuple)
        self._frames = None
        self._image_keys = None
        self._image_keys_control = None
        self._rotation_angles = None
        self._distance = None
        self._fov = None
        self._energy = None
        self._estimated_cor_frm_motor = None
        self._start_time = None
        self._end_time = None
        self._x_translations = None
        self._y_translations = None
        self._z_translations = None
        self._nexus_paths = None
        self._nexus_version = None
        self._user_nx_version = nx_version
        self.__detector_broken_vds_behavior = {
            "run_check": True,  # check vds
            "raise_error": False,  # if met an error raise it
            "log_level": logging.WARNING,  # if met an error log it with the following level
        }
        # behavior to get when check detector vds
        self._x_flipped = None
        self._y_flipped = None
        self._detector_transformations = None
        self._bliss_orginal_files = None

    def set_check_behavior(
        self, run_check=True, raise_error=False, log_level=logging.WARNING
    ):
        """
        when user require to access to scan frames NXtomoScan build them (`frames` property).
        Some check can be made during this stage to know if the scan has some broken virtual-dataset (vds) or
        if the vds is linked to more file than the system might handle.

        In this case the 'vds-check' can either raise an error or log potential issues with a specific log level
        """
        self.__detector_broken_vds_behavior = {
            "run_check": run_check,
            "raise_error": raise_error,
            "log_level": log_level,
        }

    @staticmethod
    def get_master_file(scan_path):
        if os.path.isfile(scan_path):
            master_file = scan_path
        else:
            master_file = os.path.join(scan_path, os.path.basename(scan_path))
            if os.path.exists(master_file + ".nx"):
                master_file = master_file + ".nx"
            elif os.path.exists(master_file + ".hdf5"):
                master_file = master_file + ".hdf5"
            elif os.path.exists(master_file + ".h5"):
                master_file = master_file + ".h5"
            else:
                master_file = master_file + ".nx"
        return master_file

    @docstring(TomoScanBase.clear_caches)
    def clear_caches(self) -> None:
        self._dim_1, self._dim_2 = None, None
        self._x_pixel_size = None
        self._y_pixel_size = None
        self._x_magnified_pixel_size = None
        self._y_magnified_pixel_size = None
        self._distance = None
        self._fov = None
        self._source = None
        self._energy = None
        self._x_flipped = None
        self._y_flipped = None
        self._detector_transformations = None
        super().clear_caches()

    def clear_frames_caches(self):
        self._projections_compacted = None
        self._flats = None
        self._darks = None
        self._tomo_n = None
        self._dark_n = None
        self._flat_n = None
        self._scan_range = None
        self._frames = None
        self._image_keys = None
        self._image_keys_control = None
        self._count_time = None
        self._x_flipped = None
        self._y_flipped = None
        self._detector_transformations = None
        self._x_translations = None
        self._y_translations = None
        self._z_translations = None
        self._rotation_angles = None
        super().clear_frames_caches()

    @staticmethod
    def _get_entry_at(index: int, file_path: str) -> str:
        """

        :param index:
        :param file_path:
        :return:
        """
        entries = _NXtomo.get_valid_entries(file_path)
        if len(entries) == 0:
            return None
        else:
            return entries[index]

    @staticmethod
    @deprecated(replacement="NXtomo.get_valid_entries", since_version="2.0")
    def get_valid_entries(file_path: str) -> tuple:
        """
        return the list of 'Nxtomo' entries at the root level

        :param str file_path:
        :return: list of valid Nxtomo node (ordered alphabetically)
        :rtype: tuple

        ..note: entries are sorted to insure consistency
        """
        return _NXtomo.get_valid_entries(file_path=file_path)

    @staticmethod
    @deprecated(replacement="NXtomo.node_is_nxtomo", since_version="2.0")
    def node_is_nxtomo(node: h5py.Group) -> bool:
        """check if the given h5py node is an nxtomo node or not"""
        return _NXtomo.node_is_nxtomo(node)

    @docstring(TomoScanBase.is_tomoscan_dir)
    @staticmethod
    def is_tomoscan_dir(directory: str, **kwargs) -> bool:
        if os.path.isfile(directory):
            master_file = directory
        else:
            master_file = NXtomoScan.get_master_file(scan_path=directory)
        if master_file:
            entries = _NXtomo.get_valid_entries(file_path=master_file)
            return len(entries) > 0

    @docstring(TomoScanBase.is_abort)
    def is_abort(self, **kwargs):
        # for now there is no abort definition in .hdf5
        return False

    @docstring(TomoScanBase.to_dict)
    def to_dict(self) -> dict:
        res = super().to_dict()
        res[self.DICT_PATH_KEY] = self.master_file
        res[self._DICT_ENTRY_KEY] = self.entry
        return res

    @staticmethod
    def from_dict(_dict: dict):
        scan = NXtomoScan(scan=None)
        scan.load_from_dict(_dict=_dict)
        return scan

    @docstring(TomoScanBase.load_from_dict)
    def load_from_dict(self, _dict: dict) -> TomoScanBase:
        """

        :param _dict:
        :return:
        """
        if isinstance(_dict, io.TextIOWrapper):
            data = json.load(_dict)
        else:
            data = _dict
        if not (self.DICT_TYPE_KEY in data and data[self.DICT_TYPE_KEY] == self._TYPE):
            raise ValueError("Description is not an HDF5Scan json description")
        if NXtomoScan._DICT_ENTRY_KEY not in data:
            raise ValueError("No hdf5 entry specified")

        assert self.DICT_PATH_KEY in data
        self._entry = data[self._DICT_ENTRY_KEY]
        self.master_file = self.get_master_file(data[self.DICT_PATH_KEY])

        if os.path.isdir(data[self.DICT_PATH_KEY]):
            self.path = data[self.DICT_PATH_KEY]
        else:
            self.path = os.path.dirname(data[self.DICT_PATH_KEY])
        return self

    @property
    def entry(self) -> str:
        return self._entry

    @property
    def nexus_version(self):
        if self._user_nx_version is not None:
            return self._user_nx_version
        return self._get_generic_key(
            "_nexus_version", self._NEXUS_VERSION_PATH, is_attribute=True
        )

    @nexus_version.setter
    def nexus_version(self, version):
        if not isinstance(version, float):
            raise TypeError("version expect to be a float")
        self._nexus_version = version

    @property
    def nexus_path(self):
        if self._nexus_paths is None:
            self._nexus_paths = _get_nexus_paths(self.nexus_version)
        return self._nexus_paths

    @property
    @docstring(TomoScanBase.source)
    def source(self):
        if self._source is None:
            self._source = Source(
                name=self.source_name,
                type=self.source_type,
            )
        return self._source

    @property
    def title(self):
        return self._get_generic_key("_title", self.nexus_path.NAME_PATH)

    @property
    def source_name(self):
        return self._get_generic_key("_source_name", self.nexus_path.SOURCE_NAME)

    @property
    def source_type(self):
        return self._get_generic_key("_source_type", self.nexus_path.SOURCE_TYPE)

    @property
    @docstring(TomoScanBase.instrument_name)
    def instrument_name(self) -> typing.Optional[str]:
        """

        :return: instrument name
        """
        return self._get_generic_key(
            "_instrument_name", self.nexus_path.INSTRUMENT_NAME
        )

    @property
    def sequence_name(self):
        """Return the sequence name"""
        return self._get_generic_key("_name", self.nexus_path.NAME_PATH)

    @property
    @docstring(TomoScanBase.projections)
    def sample_name(self):
        return self._get_generic_key("_sample_name", self.nexus_path.SAMPLE_NAME_PATH)

    @property
    @docstring(TomoScanBase.projections)
    def group_size(self):
        return self._get_generic_key("_grp_size", self.nexus_path.GRP_SIZE_ATTR)

    @property
    @docstring(TomoScanBase.projections)
    def projections(self) -> typing.Optional[dict]:
        if self._projections is None:
            if self.frames:
                ignored_projs = []
                if self.ignore_projections is not None:
                    ignored_projs = self.ignore_projections
                proj_frames = tuple(
                    filter(
                        lambda x: (
                            x.image_key is ImageKey.PROJECTION
                            and x.index not in ignored_projs
                            and x.is_control is False
                        ),
                        self.frames,
                    )
                )
                self._projections = {}
                for proj_frame in proj_frames:
                    self._projections[proj_frame.index] = proj_frame.url
        return self._projections

    @projections.setter
    def projections(self, projections: dict):
        self._projections = projections

    def get_projections_intensity_monitor(self) -> dict:
        """return intensity monitor values for projections"""
        if self.frames:
            ignored_projs = []
            if self.ignore_projections is not None:
                ignored_projs = self.ignore_projections
            proj_frames = tuple(
                filter(
                    lambda x: (
                        x.image_key is ImageKey.PROJECTION
                        and x.index not in ignored_projs
                        and x.is_control is False
                    ),
                    self.frames,
                )
            )
            intensity_monitor = {}
            for proj_frame in proj_frames:
                intensity_monitor[proj_frame.index] = proj_frame.intensity_monitor
            return intensity_monitor
        else:
            return {}

    @property
    @docstring(TomoScanBase.alignment_projections)
    def alignment_projections(self) -> typing.Optional[dict]:
        if self._alignment_projections is None:
            if self.frames:
                proj_frames = tuple(
                    filter(
                        lambda x: x.image_key == ImageKey.PROJECTION
                        and x.is_control is True,
                        self.frames,
                    )
                )
                self._alignment_projections = {}
                for proj_frame in proj_frames:
                    self._alignment_projections[proj_frame.index] = proj_frame.url
        return self._alignment_projections

    @property
    @docstring(TomoScanBase.darks)
    def darks(self) -> typing.Optional[dict]:
        if self._darks is None:
            if self.frames:
                dark_frames = tuple(
                    filter(lambda x: x.image_key is ImageKey.DARK_FIELD, self.frames)
                )
                self._darks = {}
                for dark_frame in dark_frames:
                    self._darks[dark_frame.index] = dark_frame.url
        return self._darks

    @property
    @docstring(TomoScanBase.flats)
    def flats(self) -> typing.Optional[dict]:
        if self._flats is None:
            if self.frames:
                flat_frames = tuple(
                    filter(lambda x: x.image_key is ImageKey.FLAT_FIELD, self.frames)
                )
                self._flats = {}
                for flat_frame in flat_frames:
                    self._flats[flat_frame.index] = flat_frame.url
        return self._flats

    @docstring(TomoScanBase.update)
    def update(self) -> None:
        """update list of radio and reconstruction by parsing the scan folder"""
        if self.master_file is None or not os.path.exists(self.master_file):
            return
        self.projections = self._get_projections_url()
        # TODO: update darks and flats too

    @docstring(TomoScanBase.get_proj_angle_url)
    def _get_projections_url(self):
        if self.master_file is None or not os.path.exists(self.master_file):
            return
        frames = self.frames
        if frames is not None:
            urls = {}
            for frame in frames:
                if frame.image_key is ImageKey.PROJECTION:
                    urls[frame.index] = frame.url
            return urls
        else:
            return None

    @docstring(TomoScanBase.tomo_n)
    @property
    def tomo_n(self) -> typing.Optional[int]:
        """we are making two asumptions for computing tomo_n:
        - if a rotation = scan_range +/- EPSILON this is a return projection
        - The delta between each projections is constant
        """
        return self._get_generic_key("_tomo_n", self.nexus_path.TOMO_N_SCAN)

    @docstring(TomoScanBase.tomo_n)
    @property
    def magnification(self):
        return self._get_generic_key(
            "_magnification",
            "/".join(
                [
                    self.nexus_path.INSTRUMENT_PATH,
                    self.nexus_path.nx_instrument_paths.DETECTOR_PATH,
                    self.nexus_path.nx_detector_paths.MAGNIFICATION,
                ]
            ),
        )

    @property
    def return_projs(self) -> typing.Optional[list]:
        """ """
        frames = self.frames
        if frames:
            return_frames = list(filter(lambda x: x.is_control is True, frames))
            return return_frames
        else:
            return None

    @property
    def rotation_angle(self) -> typing.Optional[tuple]:
        cast_to_float = lambda values: [float(val) for val in values]
        return self._get_generic_key(
            "_rotation_angles",
            self.nexus_path.ROTATION_ANGLE_PATH,
            apply_function=cast_to_float,
        )

    @property
    def x_translation(self) -> typing.Optional[tuple]:
        cast_to_float = lambda values: [float(val) for val in values]
        return self._get_generic_key(
            "_x_translations",
            self.nexus_path.X_TRANS_PATH,
            apply_function=cast_to_float,
            unit=MetricSystem.METER,
        )

    @property
    def y_translation(self) -> typing.Optional[tuple]:
        cast_to_float = lambda values: [float(val) for val in values]
        return self._get_generic_key(
            "_y_translations",
            self.nexus_path.Y_TRANS_PATH,
            apply_function=cast_to_float,
            unit=MetricSystem.METER,
        )

    @property
    def z_translation(self) -> typing.Optional[tuple]:
        cast_to_float = lambda values: [float(val) for val in values]
        return self._get_generic_key(
            "_z_translations",
            self.nexus_path.Z_TRANS_PATH,
            apply_function=cast_to_float,
            unit=MetricSystem.METER,
        )

    @property
    def image_key(self) -> typing.Optional[list]:
        return self._get_generic_key("_image_keys", self.nexus_path.IMG_KEY_PATH)

    @property
    def image_key_control(self) -> typing.Optional[list]:
        return self._get_generic_key(
            "_image_keys_control", self.nexus_path.IMG_KEY_CONTROL_PATH
        )

    @property
    def count_time(self) -> typing.Optional[list]:
        return self._get_generic_key(
            "_count_time",
            self.nexus_path.EXPOSURE_TIME_PATH,
            unit=timesystem.TimeSystem.SECOND,
        )

    @property
    @deprecated(replacement="count_time", since_version="1.0.0")
    def exposure_time(self) -> typing.Optional[list]:
        return self.count_time

    @property
    def electric_current(self) -> typing.Optional[list]:
        return self._get_generic_key(
            "_electric_current",
            self.nexus_path.ELECTRIC_CURRENT_PATH,
            unit=electriccurrentsystem.ElectricCurrentSystem.AMPERE,
        )

    @property
    def x_flipped(self) -> bool:
        deprecated_warning(
            type_="property",
            name="y_flipped",
            replacement="get_detector_transformations",
            since_version="1.3",
        )
        if self.nexus_path.nx_detector_paths.Y_FLIPPED is not None:
            return self._get_generic_key(
                "_x_flipped",
                "/".join(
                    [
                        self.nexus_path.INSTRUMENT_PATH,
                        self.nexus_path.nx_instrument_paths.DETECTOR_PATH,
                        self.nexus_path.nx_detector_paths.X_FLIPPED,
                    ]
                ),
            )
        elif self.nexus_path.nx_detector_paths.NX_TRANSFORMATIONS:
            transformations = self.detector_transformations
            if transformations is None:
                return None
            else:
                # Poor attempt of backward compatibility.
                # But if there is other transformations then there will be unmanaged.
                # it was this or raising an error. Will probably raise an Error in short term
                return len(get_lr_flip(transformations=transformations)) % 2 == 1
        else:
            return None

    @property
    def y_flipped(self) -> bool:
        deprecated_warning(
            type_="property",
            name="y_flipped",
            replacement="get_detector_transformations",
            since_version="1.3",
        )
        if self.nexus_path.nx_detector_paths.Y_FLIPPED is not None:
            # TODO: deprecate ? but replace it by what ?
            return self._get_generic_key(
                "_y_flipped",
                "/".join(
                    [
                        self.nexus_path.INSTRUMENT_PATH,
                        self.nexus_path.nx_instrument_paths.DETECTOR_PATH,
                        self.nexus_path.nx_detector_paths.Y_FLIPPED,
                    ]
                ),
            )
        elif self.nexus_path.nx_detector_paths.NX_TRANSFORMATIONS:
            transformations = self.detector_transformations
            if transformations is None:
                return None
            else:
                # Poor attempt of backward compatibility.
                # But if there is other transformations then there will be unmanaged.
                # it was this or raising an error. Will probably raise an Error in short term
                return len(get_ud_flip(transformations=transformations)) % 2 == 1
        else:
            return None

    @property
    def detector_transformations(self) -> typing.Optional[tuple]:
        """
        return tuple of `Transformation` affecting the NXdetector
        """
        if self._detector_transformations is None:
            det_transformations = []
            # bacward compatiblity, try to find x_flipped and y_flipped and convert it to `transformation`
            if (self.nexus_version or LATEST_NXTOMO_VERSION) < 1.3:
                try:
                    with HDF5File(self.master_file, "r") as h5_file:
                        x_flipped = h5_file.get(
                            "/".join(
                                [
                                    self._entry,
                                    self.nexus_path.INSTRUMENT_PATH,
                                    self.nexus_path.nx_instrument_paths.DETECTOR_PATH,
                                    self.nexus_path.nx_detector_paths.X_FLIPPED,
                                ]
                            )
                        )
                        y_flipped = h5_file.get(
                            "/".join(
                                [
                                    self._entry,
                                    self.nexus_path.INSTRUMENT_PATH,
                                    self.nexus_path.nx_instrument_paths.DETECTOR_PATH,
                                    self.nexus_path.nx_detector_paths.Y_FLIPPED,
                                ]
                            )
                        )
                except (OSError, KeyError):
                    _logger.info(
                        f"Failed to find any track of detector flip (nexus version is {self.nexus_version})"
                    )
                    return None
                else:
                    det_transformations = []
                    prev_trans = None
                    if x_flipped is True:
                        prev_trans = LRDetTransformation(depends_on=prev_trans)
                        det_transformations.append(
                            prev_trans,
                        )
                    if y_flipped is True:
                        prev_trans = UDDetTransformation(depends_on=prev_trans)
                        det_transformations.append(
                            prev_trans,
                        )

            else:
                transformations = NXdetector.load_transformations(
                    file_path=self.master_file,
                    data_path="/".join(
                        [
                            self.entry,
                            self.nexus_path.INSTRUMENT_PATH,
                            self.nexus_path.nx_instrument_paths.DETECTOR_PATH,
                        ]
                    ),
                    nexus_version=self.nexus_version,
                )
                if transformations is not None:
                    det_transformations = transformations.transformations
            self._detector_transformations = tuple(det_transformations)
        return tuple(self._detector_transformations)

    @docstring(TomoScanBase)
    def get_bounding_box(self, axis: typing.Union[str, int] = None) -> tuple:
        """
        Return the bounding box covered by the scan (only take into account the projections).
        axis is expected to be in (0, 1, 2) or (x==0, y==1, z==2)

        :note: current pixel size is given with magnification. To move back to sample space (x_translation, y_translation, z_translation)
               we need to `unmagnified` this is size
        """
        if axis is None:
            x_bb = self.get_bounding_box(axis="x")
            y_bb = self.get_bounding_box(axis="y")
            z_bb = self.get_bounding_box(axis="z")
            return BoundingBox3D(
                (z_bb.min, y_bb.min, x_bb.min),
                (z_bb.max, y_bb.max, x_bb.max),
            )

        if axis == 0:
            axis = "z"
        elif axis == 1:
            axis = "y"
        elif axis == 2:
            axis = "x"
        if axis not in ("x", "y", "z"):
            raise ValueError(
                f"Axis is expected to be in ('x', 'y', 'z', 0, 1, 2). Got {axis}."
            )

        if axis == "x":
            translations = self.x_translation
            default_pixel_size = self.x_pixel_size
            n_pixel = self.dim_1
        elif axis == "y":
            translations = self.y_translation
            default_pixel_size = self.y_pixel_size
            n_pixel = self.dim_2
        elif axis == "z":
            translations = self.z_translation
            default_pixel_size = self.y_pixel_size
            n_pixel = self.dim_2
        else:
            raise ValueError(
                f"Axis is expected to be in ('x', 'y', 'z', 0, 1, 2). Got {axis}."
            )

        if translations is None or len(translations) == 0:
            raise ValueError(f"Unable to find translation for axis {axis}")
        translations = numpy.asarray(translations)
        # TODO: might need to filter only the projection one ?
        filetered_translation_for_proj = translations[
            self.image_key_control == ImageKey.PROJECTION.value
        ]
        min_axis_translation = filetered_translation_for_proj.min()
        max_axis_translation = filetered_translation_for_proj.max()
        if default_pixel_size is None:
            raise ValueError(f"Unable to find pixel size for axis {axis}")
        if n_pixel is None:
            raise ValueError(f"Unable to find number of pixel for axis {axis}")

        min_pos_in_meter = min_axis_translation - (n_pixel / 2.0 * default_pixel_size)
        max_pos_in_meter = max_axis_translation + (n_pixel / 2.0 * default_pixel_size)
        return BoundingBox1D(min_pos_in_meter, max_pos_in_meter)

    def _get_generic_key(
        self,
        key_name,
        path_key_name,
        unit: typing.Optional[Unit] = None,
        apply_function=None,
        is_attribute=False,
    ) -> typing.Any:
        if not isinstance(unit, (type(None), Unit)):
            raise TypeError(
                f"default_unit must be an instance of {Unit} or None. Not {type(unit)}"
            )

        if getattr(self, key_name, None) is None:
            self._check_hdf5scan_validity()
            with HDF5File(self.master_file, "r") as h5_file:
                if is_attribute and path_key_name in h5_file[self._entry].attrs:
                    attr_val = h5py_read_dataset(
                        h5_file[self._entry].attrs[path_key_name]
                    )
                    if apply_function is not None:
                        attr_val = apply_function(attr_val)
                elif not is_attribute and path_key_name in h5_file[self._entry]:
                    if unit is not None:
                        attr_val = self._get_value(
                            h5_file[self._entry][path_key_name], default_unit=unit
                        )
                    else:
                        attr_val = h5py_read_dataset(
                            h5_file[self._entry][path_key_name]
                        )
                    if apply_function is not None:
                        attr_val = apply_function(attr_val)
                else:
                    attr_val = None
            setattr(self, key_name, attr_val)
        return getattr(self, key_name)

    @docstring(TomoScanBase.dark_n)
    @property
    def dark_n(self) -> typing.Optional[int]:
        if self.darks is not None:
            return len(self.darks)
        else:
            return None

    @docstring(TomoScanBase.flat_n)
    @property
    def flat_n(self) -> typing.Optional[int]:
        splitted_flats = self.splitted_flat_serie
        if splitted_flats is not None and len(splitted_flats) > 0:
            return len(splitted_flats[0])
        else:
            return None

    @property
    def splitted_flat_serie(self) -> typing.Optional[dict]:
        """split flat according to flat indices"""
        if self.flats is None:
            return None
        else:
            flat_indexes = numpy.array(tuple(self.flats.keys()))
            # group contiguous flats (if differences of indices is > 1 then we consider this is part of another serie)
            group_indices = numpy.split(
                flat_indexes, numpy.where(numpy.diff(flat_indexes) != 1)[0] + 1
            )
            # previous function can return some 'empty' arrays. Let's filter them
            contiguous_indices_groups = filter(
                lambda array: len(array) > 0,
                group_indices,
            )
            #  lets recreate the dict of flats with index as key and orginal value
            series = list()
            for contiguous_indices in contiguous_indices_groups:
                series.append(
                    {
                        flat_index: self.flats[flat_index]
                        for flat_index in contiguous_indices
                    }
                )
            return tuple(series)

    @docstring(TomoScanBase.ff_interval)
    @property
    def ff_interval(self):
        raise NotImplementedError(
            "not implemented for hdf5. But we have " "acquisition sequence instead."
        )

    @docstring(TomoScanBase.scan_range)
    @property
    def scan_range(self) -> typing.Optional[int]:
        """For now scan range should return 180 or 360. We don't expect other value."""
        if (
            self._scan_range is None
            and self.master_file
            and os.path.exists(self.master_file)
            and self._entry is not None
        ):
            rotation_angle = self.rotation_angle
            if rotation_angle is not None:
                angle_range = numpy.max(rotation_angle) - numpy.min(rotation_angle)
                dist_to180 = abs(180 - angle_range)
                dist_to360 = abs(360 - angle_range)
                if dist_to180 < dist_to360:
                    self._scan_range = 180
                else:
                    self._scan_range = 360
        return self._scan_range

    @property
    def dim_1(self) -> typing.Optional[int]:
        if self._dim_1 is None:
            self._get_dim1_dim2()
        return self._dim_1

    @property
    def dim_2(self) -> typing.Optional[int]:
        if self._dim_2 is None:
            self._get_dim1_dim2()
        return self._dim_2

    @property
    def pixel_size(self) -> typing.Optional[float]:
        """return x pixel size in meter"""
        return self.x_pixel_size

    @property
    def x_pixel_size(self) -> typing.Optional[float]:
        """return x pixel size in meter"""
        return self._get_generic_key(
            "_x_pixel_size",
            self.nexus_path.X_PIXEL_SIZE_PATH,
            unit=MetricSystem.METER,
        )

    @property
    def y_pixel_size(self) -> typing.Optional[float]:
        """return y pixel size in meter"""
        return self._get_generic_key(
            "_y_pixel_size",
            self.nexus_path.Y_PIXEL_SIZE_PATH,
            unit=MetricSystem.METER,
        )

    @property
    def x_real_pixel_size(self) -> typing.Optional[float]:
        return self._get_generic_key(
            "_y_pixel_size",
            self.nexus_path.X_REAL_PIXEL_SIZE_PATH,
            unit=MetricSystem.METER,
        )

    @property
    def y_real_pixel_size(self) -> typing.Optional[float]:
        return self._get_generic_key(
            "_y_pixel_size",
            self.nexus_path.Y_REAL_PIXEL_SIZE_PATH,
            unit=MetricSystem.METER,
        )

    def _get_fov(self):
        with HDF5File(self.master_file, "r", swmr=True, libver="latest") as h5_file:
            if self.nexus_path.FOV_PATH in h5_file[self._entry]:
                fov = h5py_read_dataset(h5_file[self._entry][self.nexus_path.FOV_PATH])
                return FOV.from_value(fov)
            else:
                return None

    def _get_dim1_dim2(self):
        if self.master_file and os.path.exists(self.master_file):
            if self.projections is not None:
                if len(self.projections) > 0:
                    url = list(self.projections.values())[0]
                    try:
                        with HDF5File(url.file_path(), mode="r") as h5s:
                            self._dim_2, self._dim_1 = h5s[url.data_path()].shape[-2:]
                    except Exception:
                        self._dim_2, self._dim_1 = get_data(
                            list(self.projections.values())[0]
                        ).shape

    @property
    def distance(self) -> typing.Optional[float]:
        """return sample detector distance in meter"""
        return self._get_generic_key(
            "_distance",
            self.nexus_path.DISTANCE_PATH,
            unit=MetricSystem.METER,
        )

    @property
    @docstring(TomoScanBase.field_of_view)
    def field_of_view(self):
        if self._fov is None and self.master_file and os.path.exists(self.master_file):
            self._fov = self._get_fov()
        return self._fov

    @property
    @docstring(TomoScanBase.estimated_cor_frm_motor)
    def estimated_cor_frm_motor(self):
        cast_to_float = lambda x: float(x)
        return self._get_generic_key(
            "_estimated_cor_frm_motor",
            self.nexus_path.ESTIMATED_COR_FRM_MOTOR_PATH,
            apply_function=cast_to_float,
        )

    @property
    def energy(self) -> typing.Optional[float]:
        """energy in keV"""
        energy_si = self._get_generic_key(
            "_energy",
            self.nexus_path.ENERGY_PATH,
            unit=energysystem.EnergySI.KILOELECTRONVOLT,
        )
        if energy_si is None:
            return None
        else:
            # has for energy we do an exception we don't use SI but kev
            energy_kev = energy_si / energysystem.EnergySI.KILOELECTRONVOLT.value
            return energy_kev

    @property
    def start_time(self):
        return self._get_generic_key("_start_time", self.nexus_path.START_TIME_PATH)

    @property
    def end_time(self):
        return self._get_generic_key("_end_time", self.nexus_path.END_TIME_PATH)

    @property
    def intensity_monitor(self):
        return self._get_generic_key(
            "_intensity_monitor", self.nexus_path.INTENSITY_MONITOR_PATH
        )

    def get_bliss_orginal_files(self) -> typing.Optional[str]:
        return self._get_generic_key("_bliss_orginal_files", "bliss_original_files")

    @property
    def frames(self) -> typing.Optional[tuple]:
        """return tuple of frames. Frames contains"""
        if self._frames is None:
            image_keys = self.image_key
            rotation_angles = self.rotation_angle
            x_translation = self.x_translation
            if x_translation is None and image_keys is not None:
                x_translation = [None] * len(image_keys)
            y_translation = self.y_translation
            if y_translation is None and image_keys is not None:
                y_translation = [None] * len(image_keys)
            z_translation = self.z_translation
            if z_translation is None and image_keys is not None:
                z_translation = [None] * len(image_keys)
            intensity_monitor = self.intensity_monitor
            if intensity_monitor is None and image_keys is not None:
                intensity_monitor = [None] * len(image_keys)
            if image_keys is not None and len(image_keys) != len(rotation_angles):
                raise ValueError(
                    "`rotation_angle` and `image_key` have incoherent size "
                    f"({len(rotation_angles)} vs {len(image_keys)}). Unable to deduce frame properties"
                )
            self._frames = []
            if self.__detector_broken_vds_behavior["run_check"]:
                self._checkDetectorDataset()

            if image_keys is None:
                # in the case there is no frame / image keys registered at all
                return self._frames

            for i_frame, rot_a, img_key, x_tr, y_tr, z_tr, i_m in zip(
                range(len(rotation_angles)),
                rotation_angles,
                image_keys,
                x_translation,
                y_translation,
                z_translation,
                intensity_monitor,
            ):
                url = DataUrl(
                    file_path=self.master_file,
                    data_slice=(i_frame),
                    data_path=self.get_detector_data_path(),
                    scheme="silx",
                )

                frame = TomoFrame(
                    index=i_frame,
                    url=url,
                    image_key=img_key,
                    rotation_angle=rot_a,
                    x_translation=x_tr,
                    y_translation=y_tr,
                    z_translation=z_tr,
                    intensity_monitor=i_m,
                )
                if self.image_key_control is not None:
                    try:
                        is_control_frame = (
                            ImageKey.from_value(
                                int(self.image_key_control[frame.index])
                            )
                            is ImageKey.ALIGNMENT
                        )
                    except Exception:
                        _logger.warning(
                            f"Unable to deduce if {frame.index} is a control frame. Consider it is not"
                        )
                        is_control_frame = False
                else:
                    # if there is no image_key_control it is ambiguous to determine which
                    # projection is a return or not. Even if an angle is repeated this wouldn't be sure at 100%
                    # so the better is to consider there is no return / alignment projection
                    is_control_frame = False
                frame.is_control = is_control_frame
                self._frames.append(frame)
            self._frames = tuple(self._frames)
        return self._frames

    @docstring(TomoScanBase.get_proj_angle_url)
    def get_proj_angle_url(self) -> typing.Optional[dict]:
        if self.frames is not None:
            res = {}
            for frame in self.frames:
                if frame.image_key is ImageKey.PROJECTION:
                    if frame.is_control is False:
                        res[frame.rotation_angle] = frame.url
                    else:
                        res[str(frame.rotation_angle) + "(1)"] = frame.url
            return res
        else:
            return None

    def _get_sinogram_ref_imp(self, line, subsampling=1):
        """call the reference implementation of get_sinogram.
        Used for unit test and insure the result is the same as get_sinogram
        """
        return TomoScanBase.get_sinogram(self, line=line, subsampling=subsampling)

    @docstring(TomoScanBase)
    def get_sinogram(
        self,
        line,
        subsampling=1,
        norm_method: typing.Optional[str] = None,
        **kwargs,
    ) -> numpy.array:
        if (
            len(self.projections) is not None
            and self.dim_2 is not None
            and line > self.dim_2
        ) or line < 0:
            raise ValueError(f"requested line {line} is not in the scan")

        if not isinstance(subsampling, int):
            raise TypeError("subsampling expected to be an int")
        if subsampling <= 0:
            raise ValueError("subsampling expected to be higher than 1")

        if self.projections is not None:
            # get the z line
            with HDF5File(self.master_file, mode="r") as h5f:
                raw_sinogram = h5f[self.get_detector_data_path()][:, line, :]

            assert raw_sinogram.ndim == 2
            ignored_projs = []
            if self.ignore_projections is not None:
                ignored_projs = self.ignore_projections

            def is_pure_projection(frame: TomoFrame):
                return (
                    frame.image_key is ImageKey.PROJECTION
                    and not frame.is_control
                    and frame.index not in ignored_projs
                )

            is_projection_array = numpy.array(
                [is_pure_projection(frame) for frame in self.frames]
            )
            # TODO: simplify & reduce with filter or map ?
            proj_indexes = []
            for x, y in zip(self.frames, is_projection_array):
                if bool(y) is True:
                    proj_indexes.append(x.index)

            raw_sinogram = raw_sinogram[is_projection_array, :]
            assert len(raw_sinogram) == len(
                proj_indexes
            ), "expect to get project indexes of the sinogram"
            assert raw_sinogram.ndim == 2, "sinogram is expected to be 2D"
            # now apply flat field correction on each line
            res = []
            for z_frame_raw_sino, proj_index in zip(raw_sinogram, proj_indexes):
                assert z_frame_raw_sino.ndim == 1
                line_corrected = self.flat_field_correction(
                    projs=(z_frame_raw_sino,),
                    proj_indexes=[
                        proj_index,
                    ],
                    line=line,
                )[0]
                assert isinstance(line_corrected, numpy.ndarray)
                assert line_corrected.ndim == 1
                res.append(line_corrected)
            sinogram = numpy.array(res)
            assert sinogram.ndim == 2
            # apply subsampling (could be speed up but not sure this is useful
            # compare to complexity that we would need to had
            return self._apply_sino_norm(
                sinogram[::subsampling].copy(),
                line=line,
                norm_method=norm_method,
                **kwargs,
            )
        else:
            return None

    def get_detector_data_path(self) -> str:
        return self.entry + "/instrument/detector/data"

    @property
    def projections_compacted(self):
        """
        Return a compacted view of projection frames.

        :return: Dictionary where the key is a list of indices, and the value
            is the corresponding `silx.io.url.DataUrl` with merged data_slice
        :rtype: dict
        """
        if self._projections_compacted is None:
            self._projections_compacted = get_compacted_dataslices(self.projections)
        return self._projections_compacted

    def __str__(self):
        file_id = os.sep.join(os.path.abspath(self.master_file).split(os.sep)[-3:])
        return f"NXtomo scan(master_file: {file_id}, entry: {self.entry})"

    @staticmethod
    def _get_value(node: h5py.Group, default_unit: Unit):
        """convert the value contained in the node to the adapted unit.
        Unit can be defined in on of the group attributes. It it is the case
        will pick this unit, otherwise will use the default unit
        """
        if not isinstance(default_unit, Unit):
            raise TypeError(
                f"default_unit must be an instance of {Unit}. Not {type(default_unit)}"
            )
        value = h5py_read_dataset(node)
        if "unit" in node.attrs:
            unit = node.attrs["unit"]
        elif "units" in node.attrs:
            unit = node.attrs["units"]
        else:
            unit = default_unit
        # handle Diamond dataset where unit is stored as bytes
        if hasattr(unit, "decode"):
            unit = unit.decode()
        return value * default_unit.from_value(unit).value

    def _check_hdf5scan_validity(self):
        """some basic check to make sure both the master file (.nx file) and requested entry (data_path) exists"""
        if self.master_file is None:
            raise ValueError("No master file provided")
        if self.entry is None:
            raise ValueError("No entry provided")
        with HDF5File(self.master_file, "r") as h5_file:
            if self._entry not in h5_file:
                raise ValueError(
                    f"Given entry {self._entry} is not in the master file {self.master_file}"
                )

    def _checkDetectorDataset(self):
        """check that detector dataset is valid - especially in the case this is a vds"""

        if os.getenv("TOMOTOOLS_SKIP_DET_CHECK", "0").lower() in (
            "1",
            "y",
        ):
            _logger.warning(
                " skipping virtual datasets check because of SKIP_TOMOSCAN_CHECK"
            )
            return

        url = DataUrl(
            file_path=os.path.abspath(self.master_file),
            data_path=self.get_detector_data_path(),
        )
        _, unique_files = dataset_has_broken_vds(
            url=url,
            raise_error=self.__detector_broken_vds_behavior["raise_error"],
            log_level=self.__detector_broken_vds_behavior["log_level"],
            return_unique_files=True,
        )
        check_possible_issue_with_rlimit(
            url=url,
            raise_error=self.__detector_broken_vds_behavior["raise_error"],
            log_level=self.__detector_broken_vds_behavior["log_level"],
            unique_files=unique_files,
        )

    def get_flat_expected_location(self):
        return DataUrl(
            file_path=self.master_file,
            data_path=_get_nexus_paths(self.nexus_version).PROJ_PATH,
        ).path()

    def get_dark_expected_location(self):
        return DataUrl(
            file_path=self.master_file,
            data_path=_get_nexus_paths(self.nexus_version).PROJ_PATH,
        ).path()

    def get_projection_expected_location(self):
        return DataUrl(
            file_path=self.master_file,
            data_path=_get_nexus_paths(self.nexus_version).PROJ_PATH,
        ).path()

    def get_energy_expected_location(self):
        return DataUrl(
            file_path=self.master_file,
            data_path=_get_nexus_paths(self.nexus_version).ENERGY_PATH,
        ).path()

    def get_distance_expected_location(self):
        return DataUrl(
            file_path=self.master_file,
            data_path=_get_nexus_paths(self.nexus_version).ENERGY_PATH,
        ).path()

    def get_pixel_size_expected_location(self):
        return DataUrl(
            file_path=self.master_file,
            data_path=_get_nexus_paths(self.nexus_version).X_PIXEL_SIZE_PATH,
        ).path()

    @docstring(TomoScanBase.get_relative_file)
    def get_relative_file(
        self, file_name: str, with_dataset_prefix=True
    ) -> typing.Optional[str]:
        if self.path is not None:
            if with_dataset_prefix:
                basename = self.get_dataset_basename()
                basename = "_".join((basename, file_name))
                return os.path.join(self.path, basename)
            else:
                return os.path.join(self.path, file_name)
        else:
            return None

    def get_dataset_basename(self) -> str:
        basename, _ = os.path.splitext(self.master_file)
        return os.path.basename(basename)

    @docstring(TomoScanBase)
    def save_reduced_darks(
        self,
        darks: dict,
        output_urls: tuple = REDUCED_DARKS_DATAURLS,
        darks_infos=None,
        metadata_output_urls=REDUCED_DARKS_METADATAURLS,
        overwrite: bool = False,
    ):
        """
        Dump computed dark (median / mean...) into files
        """
        super().save_reduced_darks(
            darks=darks,
            output_urls=output_urls,
            darks_infos=darks_infos,
            metadata_output_urls=metadata_output_urls,
            overwrite=overwrite,
        )

    @docstring(TomoScanBase)
    def load_reduced_darks(
        self,
        inputs_urls: tuple = REDUCED_DARKS_DATAURLS,
        metadata_input_urls=REDUCED_DARKS_METADATAURLS,
        return_as_url: bool = False,
        return_info: bool = False,
    ) -> dict:
        """
        load computed dark (median / mean...) into files
        """
        return super().load_reduced_darks(
            inputs_urls=inputs_urls,
            metadata_input_urls=metadata_input_urls,
            return_as_url=return_as_url,
            return_info=return_info,
        )

    @docstring(TomoScanBase)
    def save_reduced_flats(
        self,
        flats: dict,
        output_urls: tuple = REDUCED_FLATS_DATAURLS,
        flats_infos=None,
        metadata_output_urls: tuple = REDUCED_FLATS_METADATAURLS,
        overwrite: bool = False,
    ) -> dict:
        """
        Dump computed flats (median / mean...) into files
        """
        super().save_reduced_flats(
            flats=flats,
            metadata_output_urls=metadata_output_urls,
            output_urls=output_urls,
            flats_infos=flats_infos,
            overwrite=overwrite,
        )

    @docstring(TomoScanBase)
    def load_reduced_flats(
        self,
        inputs_urls: tuple = REDUCED_FLATS_DATAURLS,
        metadata_input_urls=REDUCED_FLATS_METADATAURLS,
        return_as_url: bool = False,
        return_info: bool = False,
    ) -> dict:
        """
        load computed dark (median / mean...) into files
        """
        return super().load_reduced_flats(
            inputs_urls=inputs_urls,
            metadata_input_urls=metadata_input_urls,
            return_as_url=return_as_url,
            return_info=return_info,
        )

    @docstring(TomoScanBase.compute_reduced_flats)
    def compute_reduced_flats(
        self,
        reduced_method="median",
        overwrite=True,
        output_dtype=numpy.float32,
        return_info: bool = False,
    ):
        return super().compute_reduced_flats(
            reduced_method=reduced_method,
            overwrite=overwrite,
            output_dtype=output_dtype,
            return_info=return_info,
        )

    @docstring(TomoScanBase.compute_reduced_flats)
    def compute_reduced_darks(
        self,
        reduced_method="mean",
        overwrite=True,
        output_dtype=numpy.float32,
        return_info: bool = False,
    ):
        return super().compute_reduced_darks(
            reduced_method=reduced_method,
            overwrite=overwrite,
            output_dtype=output_dtype,
            return_info=return_info,
        )

    @staticmethod
    @docstring(TomoScanBase)
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, NXtomoScanIdentifier):
            raise TypeError(
                f"identifier should be an instance of {NXtomoScanIdentifier}"
            )
        return NXtomoScan(scan=identifier.file_path, entry=identifier.data_path)

    @docstring(TomoScanBase)
    def get_identifier(self) -> ScanIdentifier:
        return NXtomoScanIdentifier(
            object=self, hdf5_file=self.master_file, entry=self.entry
        )

    def build_icat_metadata(self) -> dict:
        """
        build icat metadata dictionary filling NXtomo definition following icat definition: https://gitlab.esrf.fr/icat/hdf5-master-config/-/blob/88a975039694d5dba60e240b7bf46c22d34065a0/hdf5_cfg.xml
        """
        UNKNOW_VALUE = "unknow"

        def get_z_delta_and_z_start():
            if self.z_translation is None or len(self.z_translation) == 0:
                return UNKNOW_VALUE, UNKNOW_VALUE
            else:
                projection_z = numpy.array(self.z_translation)[
                    self.image_key_control == ImageKey.PROJECTION.value
                ]
                unique_z = numpy.unique(projection_z)
                if len(unique_z) > 1:
                    delta = unique_z[1] - unique_z[0]
                elif len(projection_z) > 0:
                    # in case z doesn't evolve with time
                    return 0, projection_z[0]
                else:
                    delta = UNKNOW_VALUE
                return unique_z[0], delta

        def get_exposure_time():
            if self.count_time is None or len(self.count_time) == 0:
                return UNKNOW_VALUE
            else:
                proj_count_time = self.count_time[
                    self.image_key_control == ImageKey.PROJECTION.value
                ]
                return proj_count_time[0] if len(proj_count_time) > 0 else UNKNOW_VALUE

        def get_pixel_size_micron():
            # icat expect the pixel size to be in micron...
            if self.pixel_size is None:
                return UNKNOW_VALUE
            else:
                return self.pixel_size / MetricSystem.MICROMETER.value

        def get_start_angle():
            if self.rotation_angle is None or len(self.rotation_angle) == 0:
                return UNKNOW_VALUE
            else:
                projection_rot_angle = numpy.array(self.rotation_angle)[
                    self.image_key_control == ImageKey.PROJECTION.value
                ]
                return (
                    projection_rot_angle[0]
                    if len(projection_rot_angle) > 0
                    else UNKNOW_VALUE
                )

        z_start, z_delta = get_z_delta_and_z_start()
        # If the information is not here avoid publishing it
        metadata = {
            "definition": "TOMO",
            "TOMO_ZStart": z_start,
            "TOMO_ZDelta": z_delta,
            "TOMO_accExposureTime": get_exposure_time(),
            # "sourceSampleDistance": , # source / sample distance
            "TOMO_pixelSize": get_pixel_size_micron(),
            "TOMO_start_angle": get_start_angle(),
            "TOMO_x_pixel_n": self.dim_1 if self.dim_1 is not None else UNKNOW_VALUE,
            "TOMO_y_pixel_n": self.dim_2 if self.dim_2 is not None else UNKNOW_VALUE,
        }
        if self.sample_name is not None:
            metadata["Sample_name"] = self.sample_name
            metadata["sample_name"] = self.sample_name
        if self.scan_range is not None:
            metadata["TOMO_scanRange"] = self.scan_range
        if self.field_of_view is not None:
            metadata["TOMO_halfAcquisition"] = str(self.field_of_view) is FOV.HALF
        if self.get_distance() is not None:
            metadata["TOMO_detectorDistance"] = (
                self.get_distance() / MetricSystem.MICROMETER.value
            )
        if self.flat_n is not None:
            metadata["TOMO_refN"] = self.flat_n
        if self.dark_n is not None:
            metadata["TOMO_darkN"] = self.dark_n
        if self.projections is not None:
            metadata["TOMO_projN"] = len(self.projections)

        return metadata


class HDF5XRD3DScan(NXtomoScan):
    """
    Class used to read nexus file representing a 3D-XRD acquisition.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._rocking = None
        self._base_tilt = None

    @property
    def rocking(self) -> typing.Optional[tuple]:
        if self._rocking is None:
            self._check_hdf5scan_validity()
            with HDF5File(self.master_file, "r") as h5_file:
                _rocking = h5py_read_dataset(
                    h5_file[self._entry][self.nexus_path.ROCKING_PATH]
                )
                # cast in float
                self._rocking = tuple([float(r) for r in _rocking])
        return self._rocking

    @property
    def base_tilt(self) -> typing.Optional[tuple]:
        if self._base_tilt is None:
            self._check_hdf5scan_validity()
            with HDF5File(self.master_file, "r") as h5_file:
                _base_tilt = h5py_read_dataset(
                    h5_file[self._entry][self.nexus_path.BASE_TILT_PATH]
                )
                # cast in float
                self._base_tilt = tuple([float(bt) for bt in _base_tilt])
        return self._base_tilt

    @property
    def frames(self) -> typing.Optional[tuple]:
        """return tuple of frames. Frames contains"""
        if self._frames is None:
            image_keys = self.image_key
            rotation_angles = self.rotation_angle
            x_translation = self.x_translation
            if x_translation is None and image_keys is not None:
                x_translation = [None] * len(image_keys)
            y_translation = self.y_translation
            if y_translation is None and image_keys is not None:
                y_translation = [None] * len(image_keys)
            z_translation = self.z_translation
            if z_translation is None and image_keys is not None:
                z_translation = [None] * len(image_keys)
            rocking = self.rocking
            if rocking is None and image_keys is not None:
                rocking = [None] * len(image_keys)
            base_tilt = self.base_tilt
            if base_tilt is None and image_keys is not None:
                base_tilt = [None] * len(image_keys)
            if image_keys is not None and len(image_keys) != len(rotation_angles):
                raise ValueError(
                    "`rotation_angle` and `image_key` have incoherent size "
                    f"({len(rotation_angles)} vs {len(image_keys)}). Unable to deduce frame properties"
                )
            self._frames = []

            for i_frame, rot_a, img_key, x_tr, y_tr, z_tr, rck, bt in zip(
                range(len(rotation_angles)),
                rotation_angles,
                image_keys,
                x_translation,
                y_translation,
                z_translation,
                rocking,
                base_tilt,
            ):
                url = DataUrl(
                    file_path=self.master_file,
                    data_slice=(i_frame),
                    data_path=self.get_detector_data_path(),
                    scheme="silx",
                )

                frame = XRD3DFrame(
                    index=i_frame,
                    url=url,
                    image_key=img_key,
                    rotation_angle=rot_a,
                    x_translation=x_tr,
                    y_translation=y_tr,
                    z_translation=z_tr,
                    rocking=rck,
                    base_tilt=bt,
                )
                if self.image_key_control is not None:
                    try:
                        is_control_frame = (
                            ImageKey.from_value(
                                int(
                                    self.image_key_control[  # pylint: disable=E1136  I don't know why this error is raised. I guess he think it can be None ?
                                        frame.index
                                    ]
                                )
                            )
                            is ImageKey.ALIGNMENT
                        )
                    except Exception:
                        _logger.warning(
                            f"Unable to deduce if {frame.index} is a control frame. Consider it is not"
                        )
                        is_control_frame = False
                else:
                    # if there is no image_key_control then we consider it is
                    # unsafe to determine what is a return / alignment frame
                    is_control_frame = False
                frame._is_control_frame = is_control_frame
                self._frames.append(frame)
            self._frames = tuple(self._frames)
        return self._frames


class TomoFrame:
    """class to store all metadata information of a NXTomo frame"""

    def __init__(
        self,
        index: int,
        url: typing.Optional[DataUrl] = None,
        image_key: typing.Union[None, ImageKey, int] = None,
        rotation_angle: typing.Optional[float] = None,
        is_control_proj: bool = False,
        x_translation: typing.Optional[float] = None,
        y_translation: typing.Optional[float] = None,
        z_translation: typing.Optional[float] = None,
        intensity_monitor: typing.Optional[float] = None,
    ):
        assert type(index) is int
        self._index = index
        if image_key is not None:
            self._image_key = ImageKey.from_value(image_key)
        else:
            self._image_key = None
        self._rotation_angle = rotation_angle
        self._url = url
        self._is_control_frame = is_control_proj
        self._data = None
        self._x_translation = x_translation
        self._y_translation = y_translation
        self._z_translation = z_translation
        self._intensity_monitor = intensity_monitor

    @property
    def index(self) -> int:
        return self._index

    @property
    def image_key(self) -> ImageKey:
        return self._image_key

    @image_key.setter
    def image_key(self, image_key: ImageKey) -> None:
        if not isinstance(image_key, ImageKey):
            raise TypeError(f"{image_key} is expected to be an instance of {ImageKey}")
        self._image_key = image_key

    @property
    def rotation_angle(self) -> float:
        return self._rotation_angle

    @rotation_angle.setter
    def rotation_angle(self, angle: float) -> None:
        self._rotation_angle = angle

    @property
    def url(self) -> DataUrl:
        return self._url

    @property
    def is_control(self) -> bool:
        return self._is_control_frame

    @property
    def x_translation(self):
        return self._x_translation

    @property
    def y_translation(self):
        return self._y_translation

    @property
    def z_translation(self):
        return self._z_translation

    @property
    def intensity_monitor(self):
        return self._intensity_monitor

    @is_control.setter
    def is_control(self, is_return: bool):
        self._is_control_frame = is_return

    def __str__(self):
        return (
            "Frame {index},: image_key: {image_key},"
            "is_control: {is_control},"
            "rotation_angle: {rotation_angle},"
            "x_translation: {x_translation},"
            "y_translation: {y_translation},"
            "z_translation: {z_translation},"
            "url: {url}".format(
                index=self.index,
                image_key=self.image_key,
                is_control=self.is_control,
                rotation_angle=self.rotation_angle,
                url=self.url.path(),
                x_translation=self.x_translation,
                y_translation=self.y_translation,
                z_translation=self.z_translation,
            )
        )


class XRD3DFrame(TomoFrame):
    """class to store all metadata information of a 3d-xrd nexus frame"""

    def __init__(
        self,
        index: int,
        url: typing.Optional[DataUrl] = None,
        image_key: typing.Union[ImageKey, int] = None,
        rotation_angle: typing.Optional[float] = None,
        is_control_proj: bool = False,
        x_translation: typing.Optional[float] = None,
        y_translation: typing.Optional[float] = None,
        z_translation: typing.Optional[float] = None,
        rocking: typing.Optional[float] = None,
        base_tilt: typing.Optional[float] = None,
    ):
        super().__init__(
            index=index,
            url=url,
            image_key=image_key,
            rotation_angle=rotation_angle,
            is_control_proj=is_control_proj,
            x_translation=x_translation,
            y_translation=y_translation,
            z_translation=z_translation,
        )
        self._rocking = rocking
        self._base_tilt = base_tilt

    @property
    def rocking(self) -> typing.Optional[float]:
        return self._rocking

    @property
    def base_tilt(self) -> typing.Optional[float]:
        return self._base_tilt

    def __str__(self):
        p_str = super(XRD3DFrame, self).__str__()
        p_str += f"rocking: {self.rocking}, base-tilt: {self.base_tilt}"
        return p_str


class HDF5TomoScan(NXtomoScan):
    def __init__(
        self,
        scan: str,
        entry: str = None,
        index: typing.Optional[int] = 0,
        ignore_projections: typing.Optional[typing.Iterable] = None,
        nx_version=None,
    ):
        deprecated_warning(
            type_="class",
            name="tomwer.core.scan.hdf5scan.HDF5TomoScan",
            replacement="tomwer.core.scan.nxtomo.NXtomoScan",
            since_version="2.0",
            reason="coherence",
        )
        super().__init__(scan, entry, index, ignore_projections, nx_version)



from silx.utils.enum import Enum as _Enum


class Method(_Enum):
    NONE = "none"
    SUBTRACTION = "subtraction"
    DIVISION = "division"
    CHEBYSHEV = "chebyshev"
    LSQR_SPLINE = "lsqr spline"



import logging
import os
from collections.abc import Mapping

import h5py
import numpy

from tomoscan.esrf.volume import HDF5Volume
from tomoscan.io import HDF5File
from tomoscan.utils.hdf5 import DatasetReader
from tomoscan.volumebase import VolumeBase

_logger = logging.getLogger(__name__)


def concatenate(output_volume: VolumeBase, volumes: tuple, axis: int) -> None:
    """
    Function to do 'raw' concatenation on volumes.

    This is agnostic of any metadata. So if you want to ensure about coherence of metadata (and data) you must do it yourself

    data will be concatenate in the order volumes are provided. Volumes data must be 3D. Concatenate data will be 3D and concatenation will be done
    over the axis `axis`

    concatenation will be done with a virtual dataset if input volumes and output_volume are HDF5Volume instances.

    warning: concatenation enforce writing data and metadata to disk

    :param output_volume VolumeBase: volume to create
    :param tuple volumes: tuple of VolumeBase instances
    :param int axis: axis to use for doing the concatenation. must be in 0, 1, 2
    """
    # 0. do some check
    if not isinstance(output_volume, VolumeBase):
        raise TypeError(
            f"output_volume is expected to be an instance of {VolumeBase}. {type(output_volume)} provided"
        )
    if not isinstance(axis, int):
        raise TypeError(f"axis must be an int. {type(axis)} provided")
    elif axis not in (0, 1, 2):
        raise ValueError(f"axis must be in (0, 1, 2). {axis} provided")
    if not isinstance(volumes, tuple):
        raise TypeError(f"volumes must be a tuple. {type(volumes)} provided")
    else:
        is_invalid = lambda y: not isinstance(y, VolumeBase)
        invalids = tuple(filter(is_invalid, volumes))
        if len(invalids) > 0:
            raise ValueError(f"Several non-volumes found. ({invalids})")

    from tomoscan.esrf.volume.jp2kvolume import JP2KVolume  # avoid cyclic import

    if isinstance(output_volume, JP2KVolume) and output_volume.rescale_data is True:
        _logger.warning(
            "concatenation will rescale data frame. If you want to avoid this please set output volume 'rescale_data' to False"
        )

    # 1. compute final shape
    def get_volume_shape():
        if axis == 0:
            new_shape = [0, None, None]
        elif axis == 1:
            new_shape = [None, 0, None]
        else:
            new_shape = [None, None, 0]

        for vol in volumes:
            vol_shape = vol.get_volume_shape()
            if vol_shape is None:
                raise ValueError(
                    f"Unable to find shape for volume {vol.get_identifier().to_str()}"
                )
            new_shape[axis] += vol_shape[axis]
            if axis == 0:
                if new_shape[1] is None:
                    new_shape[1], new_shape[2] = vol_shape[1], vol_shape[2]
                elif new_shape[1] != vol_shape[1] or new_shape[2] != vol_shape[2]:
                    raise ValueError("Found incoherent shapes. Unable to concatenate")
            elif axis == 1:
                if new_shape[0] is None:
                    new_shape[0], new_shape[2] = vol_shape[0], vol_shape[2]
                elif new_shape[0] != vol_shape[0] or new_shape[2] != vol_shape[2]:
                    raise ValueError("Found incoherent shapes. Unable to concatenate")
            else:
                if new_shape[0] is None:
                    new_shape[0], new_shape[1] = vol_shape[0], vol_shape[1]
                elif new_shape[0] != vol_shape[0] or new_shape[1] != vol_shape[1]:
                    raise ValueError("Found incoherent shapes. Unable to concatenate")
        return tuple(new_shape)

    final_shape = get_volume_shape()
    if final_shape is None:
        # should never be raised. Other error type is expected to be raised first
        raise RuntimeError("Unable to get final volume shape")

    # 2. Handle volume data (concatenation)

    if isinstance(output_volume, HDF5Volume) and numpy.all(
        [isinstance(vol, HDF5Volume)] for vol in volumes
    ):
        # 2.1 in the case of HDF5 we can short cut this by creating a virtual dataset. Would highly speed up processing avoid copy
        # note: in theory this could be done for any input_volume type using external dataset but we don't want to spend ages on
        # this use case for now. Some work around this (using EDf) has been done in nxtomomill for information. See https://gitlab.esrf.fr/tomotools/nxtomomill/-/merge_requests/115
        _logger.info("start creation of external dataset")
        with DatasetReader(volumes[0].data_url) as dataset:
            data_type = dataset.dtype

        with HDF5File(output_volume.data_url.file_path(), mode="a") as h5s:
            # 2.1.1 check data path
            if output_volume.data_url.data_path() in h5s:
                if output_volume.overwrite:
                    del h5s[output_volume.data_url.data_path()]
                else:
                    raise OSError(
                        f"Unable to save data to {output_volume.data_url.data_path()}. This path already exists in {output_volume.data_url.file_path()}. If you want you can ask to overwrite it (from the output volume)."
                    )
            # 2.1.2 create virtual layout
            v_layout = h5py.VirtualLayout(
                shape=final_shape,
                dtype=data_type,
            )

            # 2.1.3 create virtual source
            start_index = 0
            for volume in volumes:
                # provide relative path
                rel_file_path = os.path.relpath(
                    volume.data_url.file_path(),
                    os.path.dirname(output_volume.data_url.file_path()),
                )
                rel_file_path = "./" + rel_file_path
                data_path = volume.data_url.data_path()
                vol_shape = volume.get_volume_shape()
                vs = h5py.VirtualSource(
                    rel_file_path,
                    name=data_path,
                    shape=vol_shape,
                )
                stop_index = start_index + vol_shape[axis]
                if axis == 0:
                    v_layout[start_index:stop_index] = vs
                elif axis == 1:
                    v_layout[:, start_index:stop_index, :] = vs
                elif axis == 2:
                    v_layout[:, :, start_index:stop_index] = vs
                start_index = stop_index

            # 2.1.4 create virtual dataset
            h5s.create_virtual_dataset(
                name=output_volume.data_url.data_path(), layout=v_layout
            )

    else:
        # 2.1 default case (duplicate all input data slice by slice)
        # 2.1.1 special case of the concatenation other axis 0
        if axis == 0:

            def iter_input():
                for vol in volumes:
                    for data_slice in vol.browse_slices():
                        yield data_slice

            for frame_dumper, input_slice in zip(
                output_volume.data_file_saver_generator(
                    n_frames=final_shape[0],
                    data_url=output_volume.data_url,
                    overwrite=output_volume.overwrite,
                ),
                iter_input(),
            ):
                frame_dumper[:] = input_slice
        else:
            # 2.1.2 concatenation with data duplication over axis 1 or 2
            for i_z, frame_dumper in enumerate(
                output_volume.data_file_saver_generator(
                    n_frames=final_shape[0],
                    data_url=output_volume.data_url,
                    overwrite=output_volume.overwrite,
                )
            ):
                if axis == 1:
                    frame_dumper[:] = numpy.concatenate(
                        [vol.get_slice(axis=0, index=i_z) for vol in volumes],
                        axis=0,
                    )
                elif axis == 2:
                    frame_dumper[:] = numpy.concatenate(
                        [vol.get_slice(axis=0, index=i_z) for vol in volumes],
                        axis=1,
                    )
                else:
                    raise RuntimeError

    # 3. handle metadata
    for vol in volumes:
        if vol.metadata is None:
            try:
                vol.load_metadata(store=True)
            except Exception as e:
                _logger.error(f"fail to load metadata for {vol}. Error is {e}")
    output_volume.metadata = {}

    [update_metadata(output_volume.metadata, vol.metadata) for vol in volumes]
    output_volume.save_metadata()


def update_metadata(ddict_1: dict, ddict_2: dict) -> dict:
    """
    update metadata ddict_1 from ddict_2

    metadata are dict. And those dicts
    warning: will modify ddict_1
    """
    if not isinstance(ddict_1, dict) or not isinstance(ddict_2, dict):
        raise TypeError(f"ddict_1 and ddict_2 are expected to be instances of {dict}")
    for key, value in ddict_2.items():
        if isinstance(value, Mapping):
            ddict_1[key] = update_metadata(ddict_1.get(key, {}), value)
        else:
            ddict_1[key] = value
    return ddict_1


def rescale_data(data, new_min, new_max, data_min=None, data_max=None):
    if data_min is None:
        data_min = numpy.min(data)
    if data_max is None:
        data_max = numpy.max(data)
    return (new_max - new_min) / (data_max - data_min) * (data - data_min) + new_min



import pytest
from silx.io.url import DataUrl

from tomoscan.utils.hdf5 import DatasetReader


def test_errors_DatasetReader():
    with pytest.raises(TypeError):
        with DatasetReader("toto"):
            pass

    with pytest.raises(ValueError):
        with DatasetReader(DataUrl()):
            pass

    with pytest.raises(ValueError):
        with DatasetReader(DataUrl(file_path="test", data_path="dssad", data_slice=2)):
            pass



"""deprecated module. Use 'factory' instead"""

from tomoscan.utils.io import deprecated_warning


deprecated_warning(
    "Class",
    name="tomoscan.scanfactory.ScanFactory",
    reason="Has been moved",
    replacement="tomoscan.factory.TomoObjectFactory",
    only_once=True,
)
from tomoscan.factory import Factory as ScanFactory  # noqa F401



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2022 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "27/01/2022"


from tomoscan.esrf.identifier.folderidentifier import (
    BaseFolderAndfilePrefixIdentifierMixIn,
)
from tomoscan.identifier import ScanIdentifier, VolumeIdentifier
from tomoscan.utils import docstring


class _BaseEDFIdentifier(BaseFolderAndfilePrefixIdentifierMixIn):
    """Identifier specific to EDF TomoScan"""

    @property
    @docstring(ScanIdentifier)
    def scheme(self) -> str:
        return "edf"


class EDFTomoScanIdentifier(_BaseEDFIdentifier, ScanIdentifier):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, tomo_type=ScanIdentifier.TOMO_TYPE)

    @staticmethod
    def from_str(identifier):
        from tomoscan.esrf.scan.edfscan import EDFTomoScan

        return (
            BaseFolderAndfilePrefixIdentifierMixIn._from_str_to_single_frame_identifier(
                identifier=identifier,
                SingleFrameIdentifierClass=EDFTomoScanIdentifier,
                ObjClass=EDFTomoScan,
            )
        )


class EDFVolumeIdentifier(_BaseEDFIdentifier, VolumeIdentifier):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, tomo_type=VolumeIdentifier.TOMO_TYPE)

    @staticmethod
    def from_str(identifier):
        from tomoscan.esrf.volume.edfvolume import EDFVolume

        return (
            BaseFolderAndfilePrefixIdentifierMixIn._from_str_to_single_frame_identifier(
                identifier=identifier,
                SingleFrameIdentifierClass=EDFVolumeIdentifier,
                ObjClass=EDFVolume,
            )
        )



import numpy


class _BoundingBox:
    def __init__(self, v1, v2):
        if not numpy.isscalar(v1):
            v1 = tuple(v1)
        if not numpy.isscalar(v2):
            v2 = tuple(v2)
        self._min = min(v1, v2)
        self._max = max(v1, v2)

    @property
    def min(self) -> float:
        return self._min

    @property
    def max(self) -> float:
        return self._max

    def __str__(self):
        return f"({self.min}, {self.max})"

    def __eq__(self, other):
        if not isinstance(other, _BoundingBox):
            return False
        else:
            return self.min == other.min and self.max == other.max

    def get_overlap(self, other_bb):
        raise NotImplementedError("Base class")


class BoundingBox1D(_BoundingBox):
    def get_overlap(self, other_bb):
        if not isinstance(other_bb, BoundingBox1D):
            raise TypeError(f"Can't compare a {BoundingBox1D} with {type(other_bb)}")
        if (
            (self.max >= other_bb.min and self.min <= other_bb.max)
            or (other_bb.max >= self.min and other_bb.min <= self.max)
            or (other_bb.min <= self.min and other_bb.max >= self.max)
        ):
            return BoundingBox1D(
                max(self.min, other_bb.min), min(self.max, other_bb.max)
            )
        else:
            return None

    def __eq__(self, other):
        if isinstance(other, (tuple, list)):
            return len(other) == 2 and self.min == other[0] and self.max == other[1]
        else:
            return super().__eq__(other)

    def __hash__(self):
        return hash((self._min, self._max))


class BoundingBox3D(_BoundingBox):
    def get_overlap(self, other_bb):
        if not isinstance(other_bb, BoundingBox3D):
            raise TypeError(f"Can't compare a {BoundingBox3D} with {type(other_bb)}")
        self_bb_0 = BoundingBox1D(self.min[0], self.max[0])
        self_bb_1 = BoundingBox1D(self.min[1], self.max[1])
        self_bb_2 = BoundingBox1D(self.min[2], self.max[2])

        other_bb_0 = BoundingBox1D(other_bb.min[0], other_bb.max[0])
        other_bb_1 = BoundingBox1D(other_bb.min[1], other_bb.max[1])
        other_bb_2 = BoundingBox1D(other_bb.min[2], other_bb.max[2])

        overlap_0 = self_bb_0.get_overlap(other_bb_0)
        overlap_1 = self_bb_1.get_overlap(other_bb_1)
        overlap_2 = self_bb_2.get_overlap(other_bb_2)
        if overlap_0 is not None and overlap_1 is not None and overlap_2 is not None:
            return BoundingBox3D(
                (overlap_0.min, overlap_1.min, overlap_2.min),
                (overlap_0.max, overlap_1.max, overlap_2.max),
            )


def get_subvolume_shape(chunk, volume_shape):
    """
    Get the shape of a sub-volume to extract in a volume.

    :param chunk: tuple of slice
    :param volume_shape: tuple of int
    """
    shape = []
    for c, v in zip(chunk, volume_shape):
        start = c.start or 0
        end = c.stop or v
        if end < 0:
            end += v
        shape.append(end - start)
    return tuple(shape)
